"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A file selection GUI.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr

clr.AddReference("System.Windows.Forms")
clr.AddReference("IronPython.Wpf")

# import WPF creator and base window
import wpf
from System import Windows
import ctypes

# import settings class
# from duHast.UI import FileSelectSettings as set
from duHast.UI.file_list import get_revit_files_for_processing
from duHast.UI.Objects.file_select_settings import FileSelectionSettings


def Mbox(title, text, style):
    """
    A simple win forms message box.

    :param title: The title of the message box.
    :type title: str
    :param text: The text displayed in the message box.
    :type text: str
    :param style: An int representing the type of buttons shown.
        (https://docs.microsoft.com/en-us/dotnet/api/system.windows.forms.messageboxbuttons?view=windowsdesktop-6.0)
    :type style: int

    :return: A win form message box.
    :rtype: _type_
    """

    return ctypes.windll.user32.MessageBoxW(0, text, title, style)


# UI class
class MyWindow(Windows.Window):
    def __init__(self, xaml_full_file_name, revitFiles, settings):
        """
        Class constructor

        :param xaml_full_file_name: Fully qualified file path to wpf XAML file.
        :type xaml_full_file_name: str
        :param revitFiles: List containing path to files.
        :type revitFiles: [str]
        :param settings: A settings object.
        :type settings: :class:`.FileSelectionSettings`
        """

        wpf.LoadComponent(self, xaml_full_file_name)

        # set the settings
        if isinstance(settings, FileSelectionSettings) == False:
            raise ValueError("settings parameter must be of type FileSelectionSettings")
        self.GUIChange = True
        # populate fields
        self.selectedFiles = []
        self.revitfiles = revitFiles
        self.files.ItemsSource = revitFiles
        self.tbSourceFolder.Text = settings.input_directory
        self.tbDestinationFolder.Text = settings.output_dir
        self.tbFileType.Text = settings.revit_file_extension
        self.tbNoOfFiles.Text = str(settings.output_file_num)
        self.cbInclSubDirs.IsChecked = settings.incl_sub_dirs
        self.GUIChange = False

    def TextBox_TextChanged(self, sender, e):
        """
        Event handler for text changed in TextBox.
        """
        filter_text = (
            sender.Text.lower()
        )  # Convert to lowercase for case-insensitive filtering
        # self.FilteredFiles.Filter = lambda item: filter_text in item.name.lower()
        filtered_files = [
            file for file in self.revitfiles if filter_text in file.name.lower()
        ]
        self.files.ItemsSource = filtered_files

    def _HandleFileChange(self):
        """
        _summary_
        """

        if self.GUIChange == False:
            revitFiles = get_revit_files_for_processing(
                self.tbSourceFolder.Text,
                self.cbInclSubDirs.IsChecked,
                self.tbFileType.Text,
            )

            self.revitfiles = revitFiles
            # set new source
            self.files.ItemsSource = revitFiles

    def BtnOK(self, sender, EventArgs):
        """
        Ok button event handler.

        Gets the selected rows of files and adds them to .selectedFiles property
        Sets the dialog result value to True.

        :param sender: _description_
        :type sender: _type_
        :param EventArgs: _description_
        :type EventArgs: _type_
        """

        # get selected items
        rows = self.files.SelectedItems
        if rows != None and len(rows) > 0:
            self.selectedFiles = []
            for row in rows:
                # get the original file element
                for o_rev in self.revitfiles:
                    if o_rev.name == row.name:
                        self.selectedFiles.append(o_rev)
                        break
            # Mbox('ok',str(len(self.selectedFiles)), 1)
            self.DialogResult = True
            self.Close()
        else:
            Mbox("Attention", "No files selected", 1)

    def BtnCancel(self, sender, EventArgs):
        """
        Cancel button event handler.
        Sets the dialogue result value to False.

        :param sender: _description_
        :type sender: _type_
        :param EventArgs: _description_
        :type EventArgs: _type_
        """

        # print('cancel')
        self.DialogResult = False
        self.Close()

    def TextBoxSourcePath_TextChanged(self, sender, TextChangedEventArgs):
        """
        Source Path text box change event handler

        :param sender: _description_
        :type sender: _type_
        :param TextChangedEventArgs: _description_
        :type TextChangedEventArgs: _type_
        """

        self._HandleFileChange()

    def CheckBoxSubDir_Checked(self, sender, RoutedEventArgs):
        """
        _summary_

        :param sender: _description_
        :type sender: _type_
        :param RoutedEventArgs: _description_
        :type RoutedEventArgs: _type_
        """
        self._HandleFileChange()

    def SubDirCheckBox_Unchecked(self, sender, RoutedEventArgs):
        """
        _summary_

        :param sender: _description_
        :type sender: _type_
        :param RoutedEventArgs: _description_
        :type RoutedEventArgs: _type_
        """
        self._HandleFileChange()



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit line line patterns helper functions. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)
import System

# import common library modules
from duHast.Revit.Common import delete as rDel, parameter_get_utils as rParaGet
from duHast.Utilities.Objects import result as res


# import Autodesk
from Autodesk.Revit.DB import (
    BuiltInParameter,
    Element,
    ElementId,
    GraphicsStyleType,
    LinePatternElement,
    FilteredElementCollector,
)

# -------------------------------------------------PATTERN PROPERTIES ------------

#: pattern name
PROPERTY_PATTERN_NAME = "PatternName"
#: pattern name default value, hard coded solid line pattern name
PROPERTY_PATTERN_NAME_VALUE_DEFAULT = "Solid"
#: pattern id
PROPERTY_PATTERN_ID = "PatternId"


def get_line_pattern_from_category(cat, doc):
    """
    Returns the line pattern properties as a dictionary\
         where keys are pattern name and pattern id.

    :param cat: A category.
    :type cat: Autodesk.REvit.DB.Category
    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A dictionary.
    :rtype: dictionary {str: str, str: Autodesk.Revit.DB.ElementId}
    """

    dic_pattern = {}
    dic_pattern[PROPERTY_PATTERN_NAME] = PROPERTY_PATTERN_NAME_VALUE_DEFAULT
    dic_pattern[PROPERTY_PATTERN_ID] = pattern_id = cat.GetLinePatternId(
        GraphicsStyleType.Projection
    )
    """check for 'solid' pattern which apparently is not a pattern at all
    *The RevitAPI.chm documents says: Note that Solid is special. It isn't a line pattern at all -- 
    * it is a special code that tells drawing and export code to use solid lines rather than patterned lines. 
    * Solid is visible to the user when selecting line patterns. 
    """
    if pattern_id != LinePatternElement.GetSolidPatternId():
        # not a solid line pattern
        collector = FilteredElementCollector(doc).OfClass(LinePatternElement)
        line_pattern_element = None
        for c in collector:
            if pattern_id == c.Id:
                dic_pattern[PROPERTY_PATTERN_NAME] = Element.Name.GetValue(c)
    return dic_pattern


def get_line_pattern_from_level_element(doc, level):
    """
    Returns the line pattern properties as a dictionary\
         where keys are pattern name and pattern id.

    :param doc: Current Revit family document.
    :type doc: Current Revit family document.
    :param level: a level element
    :type level: Autodesk.Revit.DB.Level

    :return: A dictionary.
    :rtype: dictionary {str: str, str: Autodesk.Revit.DB.ElementId}
    """

    dic_pattern = {}
    dic_pattern[PROPERTY_PATTERN_NAME] = PROPERTY_PATTERN_NAME_VALUE_DEFAULT
    dic_pattern[PROPERTY_PATTERN_ID] = ElementId.InvalidElementId
    try:
        l_type_id = level.GetTypeId()
        level_type = doc.GetElement(l_type_id)
        line_pattern_id_string = rParaGet.get_built_in_parameter_value(
            level_type, BuiltInParameter.LINE_PATTERN
        )
        dic_pattern[PROPERTY_PATTERN_ID] = ElementId(int(line_pattern_id_string))
        dic_pattern[PROPERTY_PATTERN_NAME] = Element.Name.GetValue(level_type)
    except Exception as ex:
        dic_pattern[PROPERTY_PATTERN_NAME] = str(ex)
    return dic_pattern


# ------------------------------------------------ DELETE LINE PATTERNS ----------------------------------------------


def delete_line_patterns_contains(doc, contains):
    """
    Deletes all line patterns where the names contains a provided string

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param contains: Filter: pattern name needs to contain this string to be deleted.
    :type contains: str

    :return:
        Result class instance.

        - .status True if line pattern where deleted successfully. Otherwise False.
        - .message will contain delete status per pattern.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    # get all line patterns
    lps = get_all_line_patterns(doc)
    
    ids = list(
        lp.Id for lp in lps if lp.GetLinePattern().Name.Contains(contains)
    ).ToList[ElementId]()

    # is there anything to delete?
    if (ids.Count == 0):
        return_value.append_message("No line patterns found containing: {}".format(contains))
        return return_value
    
    result = rDel.delete_by_element_ids(
        doc,
        ids,
        "Deleting line patterns where name contains: {}".format(contains),
        "line patterns containing: {}".format(contains),
    )
    return result


def delete_line_pattern_starts_with(doc, starts_with):
    """
    Deletes all line patterns where the name starts with provided string.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param starts_with: Filter: pattern name needs to start with this string to be deleted.
    :type starts_with: str

    :return:
        Result class instance.

        - .status True if line pattern where deleted successfully. Otherwise False.
        - .message will contain delete status per pattern.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    # get all line patterns
    lps = get_all_line_patterns(doc)
    
    # built a .net list of element ids for deletion
    ids = list(
        lp.Id for lp in lps if lp.GetLinePattern().Name.StartsWith(starts_with)
    ).ToList[ElementId]()

    # is there anything to delete?
    if (ids.Count == 0):
        return_value.append_message("No line patterns found starting with: {}".format(starts_with))
        return return_value

    result = rDel.delete_by_element_ids(
        doc,
        ids,
        "Delete line patterns where name starts with: {}".format(starts_with),
        "line patterns starting with: {}".format(starts_with),
    )
    return result


def delete_line_patterns_without(doc, contains):
    """
    Deletes all line patterns where the name does not contain the provided string.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param contains: Filter: pattern name needs not to contain this string to be deleted.
    :type contains: str

    :return:
        Result class instance.

        - .status True if line pattern where deleted successfully. Otherwise False.
        - .message will contain delete status per pattern.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    # get all line patterns
    lps = get_all_line_patterns(doc)

    # built a .net list of element ids for deletion
    ids = list(
        lp.Id for lp in lps if not lp.GetLinePattern().Name.Contains(contains)
    ).ToList[ElementId]()

    # is there anything to delete?
    if (ids.Count == 0):
        return_value.append_message("No line patterns found starting with: {}".format(contains))
        return return_value
    
    result = rDel.delete_by_element_ids(
        doc,
        ids,
        "Delete line patterns where name does not contain: {}".format(contains),
        "line patterns without: ".format(contains),
    )
    
    return result


def get_all_line_patterns(doc):
    """
    Gets all line patterns in the model.

    :param doc: _description_
    :type doc: _type_

    :return: filtered element collector of all line pattern elements in model.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """
    return FilteredElementCollector(doc).OfClass(LinePatternElement)


def build_patterns_dictionary_by_name(doc):
    """
    Returns a dictionary where line pattern name is key, values are all ids of line patterns with the exact same name.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A dictionary where line pattern name is key, values are all ids of line patterns with the exact same name
    :rtype: dictionary(key str, value list of Autodesk.Revit.DB.ElementId)
    """

    lp_dic = {}
    lps = get_all_line_patterns(doc)
    for lp in lps:
        pattern_name = lp.GetLinePattern().Name
        if pattern_name in lp_dic:
            lp_dic[pattern_name].append(lp.Id)
        else:
            lp_dic[pattern_name] = [lp.Id]
    return lp_dic


def delete_duplicate_line_patter_names(doc):
    """
    Deletes all but the first line pattern by Id with the exact same name.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return:
        Result class instance.

        - .status True if all views where deleted. Otherwise False.
        - .message will contain deletion status.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    return_value.append_message(
        "Deletes all but the first line pattern by Id with the exact same name...start"
    )
    # get a dictionary: Key pattern name, value all ids of line patterns with the same name
    # anything where the value list is greater then 1 means duplicates of the same name...
    line_patterns = build_patterns_dictionary_by_name(doc)
    for key, value in line_patterns.items():
        if len(value) > 1:
            # keep the first one (original)
            value.remove(value[0])
            flag_delete = rDel.delete_by_element_ids(
                doc,
                value,
                "Deleting duplicate line patterns names: {}".format(key),
                "line patterns duplicates: {}".format(key),
            )
            return_value.update(flag_delete)
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit API utility functions for Rooms.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#!/usr/bin/python
# -*- coding: utf-8 -*-
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

import System
import clr

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)

# import everything from Autodesk Revit DataBase namespace (Revit API)
import Autodesk.Revit.DB as rdb

from Autodesk.Revit.DB import (
    FilteredElementCollector,
    BuiltInCategory,
    BuiltInParameter,
    ViewType,
    SpatialElementBoundaryLocation,
    SpatialElementBoundaryOptions,
)
from Autodesk.Revit.DB.Architecture import Room

from duHast.Revit.Common.parameter_get_utils import param_is_empty_or_null
from duHast.Revit.Common.Geometry.geometry import (
    UV_pt_list_from_crv_list,
    point_in_polygon,
)
from duHast.Revit.Rooms.Geometry.geometry import get_room_boundary_loops
from duHast.Revit.Common.parameter_get_utils import get_built_in_parameter_value
from duHast.Revit.Common import (
    parameter_get_utils as rParaGet,
    phases as rPhase,
)

from duHast.Utilities.utility import encode_utf8


def get_room_number(room):
    """
    Get the room number of the room
    :param room: The room to get the number of
    :type room: Room
    :return: The room number of the room
    :rtype: str
    """

    number_param = get_built_in_parameter_value(
        element=room, built_in_parameter_def=BuiltInParameter.ROOM_NUMBER
    )
    return number_param


def get_room_name(room):
    """
    Get the room name of the room
    :param room: The room to get the name of
    :type room: Room
    :return: The room name of the room
    :rtype: str
    """
    name_param = get_built_in_parameter_value(
        element=room, built_in_parameter_def=BuiltInParameter.ROOM_NAME
    )
    return name_param


def get_room_phase(rvt_doc, room):
    """
    Get the phase of the room
    :param rvt_doc: The document to get the room from
    :type rvt_doc: Document
    :param room: The room to get the phase of
    :type room: Room
    :return: The phase of the room
    :rtype: str
    """
    phase_param = get_built_in_parameter_value(
        room,
        BuiltInParameter.ROOM_PHASE,
        rParaGet.get_parameter_value_as_element_id,
    )

    phase = encode_utf8(rPhase.get_phase_name_by_id(rvt_doc, phase_param))
    return phase

def get_room_phase_object(rvt_doc, room):
    """
    Get the revit phase object of the room
    :param rvt_doc: The document to get the room from
    :type rvt_doc: Document
    :param room: The room to get the phase of
    :type room: Room
    :return: The phase of the room
    :rtype: str
    """
    phase_param = get_built_in_parameter_value(
        room,
        BuiltInParameter.ROOM_PHASE,
        rParaGet.get_parameter_value_as_element_id,
    )

    phase_object = rvt_doc.GetElement(phase_param)
    return phase_object


def get_room_level(rvt_doc, room):
    """
    Get the phase of the room
    :param rvt_doc: The document to get the room from
    :type rvt_doc: Document
    :param room: The room to get the phase of
    :type room: Room
    :return: The phase of the room
    :rtype: str
    """

    try:
        rm_level_name = rvt_doc.GetElement(room.LevelId).Name
    except Exception:
        rm_level_name = "Not placed in model"
    return rm_level_name


def get_room_num_variations(room):
    """
    Get a list of variations of room number
    e.g. 1.01, 101, 1-01
    :param room: The room to get the number of
    :type room: Room
    :return: The list of variations of room number
    :rtype: list
    """
    room_num = get_room_number(room)
    room_num_remove_fullstops = room_num.replace(".", "")
    return [room_num, room_num_remove_fullstops]


def get_room_num_name_comb(room):
    """
    Get a combination of room number and name in the format:
    '101 - Room Name'
    :param room: The room to get the number and name of
    :type room: Room
    :return: The combination of room number and name
    :rtype: str
    """
    return "{} - {}".format(get_room_number(room), get_room_name(room))


def get_rooms_in_view(document, targ_view):
    """
    Get all rooms in the view. This avoids using the FilteredElementCollector
    overload that takes a view id as an argument as this is has to generate
    graphics for all elements in the view which is very slow.
    :param document: The document to get the rooms from
    :type document: Document
    :param targ_view: The view to get the rooms from
    :type targ_view: View
    :return: The rooms in the view
    :rtype: list
    """
    view_specific_elements = (
        FilteredElementCollector(document)
        .WhereElementIsNotElementType()
        .OwnedByView(targ_view.Id)
        .ToElements()
    )
    return [el for el in view_specific_elements if el.GetType() == Room]


def get_room_from_element(phase_dictionary, elem):
    """
    Get the room associated with the element. This should be updated
    by Revit as the element moves
    :param phase_dictionary: The dictionary of phases names to phase elements
    :type phase_dictionary: dict
    :param elem: The element to get the room of
    :type elem: Element
    :return: The room associated with the element
    :rtype: Room
    """
    el_phase = elem.get_Parameter(BuiltInParameter.PHASE_CREATED)
    if el_phase != None:
        el_phase_name = el_phase.AsValueString()
        if el_phase_name != "None":
            el_phase_elem = phase_dictionary[el_phase_name]
            el_room = elem.get_Room(el_phase_elem)
            # if theres a room associated with the element return it
            return (
                el_room if el_room != None else None
            )  #'Elem Id: {} has no room'.format(str(elem.Id.IntegerValue))
        else:
            return None

    else:
        print("Elem Id: {} has no phase".format(str(elem.Id.IntegerValue)))
        return None


def check_element_is_in_room(phase_dictionary, elem, room_to_check):
    """
    Check if the element is in the room
    :param phase_dictionary: The dictionary of phases names to phase elements
    :type phase_dictionary: dict
    :param el: The element to check
    :type el: Element
    :param room_to_check: The room to check if the element is in
    :type room_to_check: Room
    :return: True if the element is in the room, False otherwise
    :rtype: bool
    """
    el_room = get_room_from_element(phase_dictionary, elem)
    # First get the room number of the element
    if el_room != None:
        # if theres a room associated with the element return the number of it
        elem_room_num = get_room_number(el_room)
        # get the number of the exemplar room to compare
        exemplar_room_num = get_room_number(room_to_check)
        # if exemplar room and element room are same num
        return True if elem_room_num == exemplar_room_num else False

    else:
        print("Elem Id: {} has no room".format(str(elem.Id.IntegerValue)))
        return False


def get_room_from_selection(rvt_doc, id_list):
    """
    Get the first Room from a selection of element ids
    :param rvt_doc: The document to get the room from
    :type rvt_doc: Document
    :param id_list: The list of element ids to check
    :type id_list: list
    :return: The room from the selection
    :rtype: Room
    """
    selected_room = None

    for elem_id in id_list:
        sel_elem = rvt_doc.GetElement(elem_id)
        if sel_elem.GetType() == Room:
            return sel_elem

    return selected_room


def check_room_on_view_level(rvt_doc, room, view):
    """
    Check if the room is on the same level as the view
    :param rvt_doc: The document to get the room from
    :type rvt_doc: Document
    :param room: The room to check
    :type room: Room
    :param view: The view to check
    :type view: View
    :return: True if the room is on the same level as the view, False otherwise
    :rtype: bool
    """
    rm_level_name = rvt_doc.GetElement(room.LevelId).Name
    view_level_name = view.get_Parameter(BuiltInParameter.PLAN_VIEW_LEVEL).AsString()
    check = rm_level_name == view_level_name
    return check


def check_rm_pt_extents(rvt_doc, room_list, target_view, target_view_crop_bx_pts):
    """
    Check if all the points from the room boundary segments are within a view crop box
    for a list of rooms
    :param rvt_doc: The document to get the room from
    :type rvt_doc: Document
    :param room_list: The list of rooms to check
    :type room_list: list
    :param target_view: The view to check
    :type target_view: View
    :param target_view_crop_bx_pts: The points of the view crop box curves
    :type target_view_crop_bx_pts: list
    :return: The rooms that are entirely within the view crop box
    :rtype: list
    """
    entire_rooms = []
    for rm in room_list:
        rm_is_on_level = check_room_on_view_level(rvt_doc, rm, target_view)

        rm_crv_list = get_room_boundary_loops(rm)

        if rm_is_on_level:
            if len(rm_crv_list) > 0:
                rm_uv_list = UV_pt_list_from_crv_list(rm_crv_list)

                pt_in_rm = []

                for rm_pt in rm_uv_list:
                    pt_in_rm.append(point_in_polygon(rm_pt, target_view_crop_bx_pts))

                if all(pt_in_rm):
                    entire_rooms.append(rm)
            else:
                return []

    return entire_rooms


def check_rm_loc_point(room_list, target_view_crop_bx_pts):
    """
    Check if the location crosshairs of a room is within a view crop box
    for a list of rooms
    :param room_list: The list of rooms to check
    :type room_list: list
    :param target_view_crop_bx_pts: The points of the view crop box curves
    :type target_view_crop_bx_pts: list
    :return: The rooms that have their location point within the view crop box
    :rtype: list
    """
    loc_pt_in = []
    for rm in room_list:
        rm_loc_pt = rm.Location.Point
        rm_loc_UV = (rm_loc_pt.X, rm_loc_pt.Y)

        if point_in_polygon(rm_loc_UV, target_view_crop_bx_pts):
            loc_pt_in.append(rm)

    return loc_pt_in


def get_only_entire_rooms(rvt_doc, rm_list, view, check_rm_edges=False):
    """
    Get only the rooms that are entirely within the view crop box
    :param rvt_doc: The document to get the room from
    :type rvt_doc: Document
    :param rm_list: The list of rooms to check
    :type rm_list: list
    :param view: The view to check
    :type view: View
    :param check_rm_edges: True: Check if the entire room boundary segments are in the view crop box
    False: Check if the location crosshair is in the view crop box
    :type check_rm_edges: bool
    :return: The rooms that are entirely within the view crop box
    :rtype: list
    """

    if view.CropBoxActive == False:
        return rm_list
    else:
        crop_reg = view.GetCropRegionShapeManager()
        crv_list = crop_reg.GetCropShape()

        if len(list(crv_list)) > 0:
            # TODO: Adaption required to handle rooms with multiple boundary segment
            # loops. Currently only handles the first loop
            view_pt_list = UV_pt_list_from_crv_list(crv_list[0])

            if check_rm_edges:
                return check_rm_pt_extents(rvt_doc, rm_list, view, view_pt_list)
            else:
                return check_rm_loc_point(rm_list, view, view_pt_list)

        else:
            return []


def get_all_rooms_in_plan_views(rvt_doc, view_list):
    """
    Get all rooms in the plan views
    :param rvt_doc: The document to get the room from
    :type rvt_doc: Document
    :param view_list: The list of views to check
    :type view_list: list
    :return: The rooms in the plan views
    :rtype: list
    """
    rooms = []

    for view in view_list:
        rooms_in_view = (
            FilteredElementCollector(rvt_doc, view.Id)
            .OfCategory(BuiltInCategory.OST_Rooms)
            .ToElements()
        )
        entire_rms = get_only_entire_rooms(rvt_doc, rooms_in_view, view)
        if len(entire_rms) > 0:
            rooms.extend(entire_rms)

    return rooms


def get_rooms_from_sheet_obj_list(rvt_doc, sheet_obj_list):
    """
    Get all rooms from the sheet object list
    :param rvt_doc: The document to get the room from
    :type rvt_doc: Document
    :param sheet_obj_list: The list of ViewSheetBaseObject objects to check
    :type sheet_obj_list: list
    :return: The rooms from the sheet object list
    :rtype: list
    """

    rms_on_sel_shts = []

    for sht_obj in sheet_obj_list:
        sht = sht_obj.sheet
        rms_on_sel_shts.extend(get_all_rooms_in_plan_views(rvt_doc, [sht]))

    return rms_on_sel_shts



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Model health report functions - annotation types.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Model health report metrics can either be displayed in a family where each parameter is assigned to a metric 
and or data can be exported to text files which can be used to visualize key metrics over time.

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr


import System

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)

from duHast.Revit.ModelHealth.Reporting.Properties.constants import (
    FAILED_TO_RETRIEVE_VALUE,
)

from duHast.Revit.Annotation.text import get_all_text_types
from duHast.Revit.Annotation.purge_unused_annotation_types import (
    get_all_unused_text_type_ids_in_model,
    get_all_unused_dim_type_ids_in_model,
    get_all_unused_arrow_type_ids_in_model,
)

from duHast.Revit.Annotation.dimensions import get_dim_types
from duHast.Revit.Annotation.arrow_heads import get_arrow_types_in_model


# ----------------------------- text -----------------------------------------------
def get_number_of_text_types(doc):
    """
    Gets the number of text types in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: Number of text types in model. On exception it will return -1
    :rtype: int
    """

    number = FAILED_TO_RETRIEVE_VALUE
    try:
        number = len(get_all_text_types(doc).ToList())
    except Exception as e:
        raise ValueError("Failed to get number of text types: {}".format(e))
    return number


def get_number_of_unused_text_types(doc):
    """
    Gets the number of unused text types in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: Number of unused text types in model. On exception it will return -1
    :rtype: int
    """

    number = FAILED_TO_RETRIEVE_VALUE
    try:
        number = len(get_all_unused_text_type_ids_in_model(doc))
    except Exception as e:
        raise ValueError("Failed to get number of unused text types: {}".format(e))
    return number


# ----------------------------- dimensions -----------------------------------------------


def get_number_of_dimension_types(doc):
    """
    Gets the number of dimension types in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: Number of dimension types in model. On exception it will return -1
    :rtype: int
    """

    number = FAILED_TO_RETRIEVE_VALUE
    try:
        number = len(get_dim_types(doc).ToList())
    except Exception as e:
        raise ValueError("Failed to get number of dimension types: {}".format(e))
    return number


def get_number_of_unused_dimension_types(doc):
    """
    Gets the number of unused dimension types in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: Number of unused dimension types in model. On exception it will return -1
    :rtype: int
    """

    number = FAILED_TO_RETRIEVE_VALUE
    try:
        number = len(get_all_unused_dim_type_ids_in_model(doc))
    except Exception as e:
        raise ValueError("Failed to get number of unused dimension types: {}".format(e))
    return number


# ----------------------------- arrow heads -----------------------------------------------


def get_number_of_arrow_head_types(doc):
    """
    Gets the number of arrow head types in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: Number of arrow head types in model. On exception it will return -1
    :rtype: int
    """

    number = FAILED_TO_RETRIEVE_VALUE
    try:
        number = len(get_arrow_types_in_model(doc))
    except Exception as e:
        raise ValueError("Failed to get number of arrow head types: {}".format(e))
    return number


def get_number_of_unused_arrow_head_types(doc):
    """
    Gets the number of unused arrow head types in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: Number of unused arrow head types in model. On exception it will return -1
    :rtype: int
    """

    number = FAILED_TO_RETRIEVE_VALUE
    try:
        number = len(get_all_unused_arrow_type_ids_in_model(doc))
    except Exception as e:
        raise ValueError(
            "Failed to get number of unused arrow head types: {}".format(e)
        )
    return number



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to purging Revit stairs and stair sub element types.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Family import purge_unused_family_types as rFamPurge
from duHast.Revit.Common import purge_utils as rPurgeUtils
from duHast.Revit.Common import parameter_get_utils as rParaGet
from duHast.Revit.Stairs import stairs as rStair
from duHast.Revit.Stairs.Utility import stairs_type_sorting as rStairSort
from duHast.Revit.Stairs import cut_marks as rStairCut
from duHast.Revit.Stairs import landings as rStairLanding
from duHast.Revit.Stairs import path as rStairPath
from duHast.Revit.Stairs import runs as rStairRun
from duHast.Revit.Stairs import stringers_carriages as rStairStringersAndCarriages

#: Built in stair family name for basic stairs
BASIC_STAIR_FAMILY_NAME = "Stair"

#: Built in stair family name for assembled stairs
ASSEMBLED_STAIR_FAMILY_NAME = "Assembled Stair"

#: Built in stair family name for precast stairs
PRECAST_STAIR_FAMILY_NAME = "Precast Stair"

#: Built in stair family name for cast in place stairs
CAST_IN_PLACE_STAIR_FAMILY_NAME = "Cast-In-Place Stair"

#: List of all Built in stair family names
BUILTIN_STAIR_TYPE_FAMILY_NAMES = [
    BASIC_STAIR_FAMILY_NAME,
    ASSEMBLED_STAIR_FAMILY_NAME,
    PRECAST_STAIR_FAMILY_NAME,
    CAST_IN_PLACE_STAIR_FAMILY_NAME,
]


def get_used_stair_type_ids(doc):
    """
    Gets all used in Stair type ids.
    Used: at least one instance of this type is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing stair types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, rStair.get_all_stair_type_ids_by_category, 1
    )
    return ids


def family_no_types_in_use(fam_type_ids, un_used_type_ids):
    """
    Compares two lists of ids. True if any id is not in un_used_type_ids.
    TODO: check for more generic list comparison and remove this function.
    :param fam_type_ids: List of family type ids to check.
    :type fam_type_ids: List of Autodesk.Revit.DB.ElementId
    :param un_used_type_ids: Reference list of ids.
    :type un_used_type_ids: List of Autodesk.Revit.DB.ElementId
    :return: True if any id from amTypeIds is not in un_used_type_ids.
    :rtype: bool
    """

    match = True
    for fam_type_id in fam_type_ids:
        if fam_type_id not in un_used_type_ids:
            match = False
            break
    return match


# -------------------------------- none in place Stair types purge -------------------------------------------------------
def get_unused_non_in_place_stair_type_ids_to_purge(doc):
    """
    Gets all unused Stair type ids for.
    Included are:
    - Stair Soffit
    - Compound Stair
    - Basic Stair
    It will therefore not return any in place family types.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing unused stair types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    # get unused type ids
    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, rStair.get_all_stair_type_ids_by_class, 0
    )
    # make sure there is at least on Stair type per system family left in model
    stair_types = rStairSort.sort_stair_types_by_family_name(doc)
    for key, value in stair_types.items():
        if key in BUILTIN_STAIR_TYPE_FAMILY_NAMES:
            if family_no_types_in_use(value, ids) == True:
                # remove one type of this system family from unused list
                ids.remove(value[0])
    return ids


# --------------------------------utility functions to get unused sub types ----------------------


def get_used_sub_type_ids_from_stair_type(doc, stair_type_id, paras):
    """
    Gets the id of types making up a stair.
    These could be stair landing types, stringer and carriage types etc.
    Types returned depend on parameter definitions past in.
    Refer to:
    - STAIR_LANDING_TYPE_PARAS,
    - STAIR_CUTMARK_TYPE_PARAS,
    - STAIR_SUPPORT_TYPE_PARAS

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param stair_type_id: An element id representing a stair type.
    :type stair_type_id: Autodesk.Revit.DB.ElementId
    :param paras: Parameters containing a type making up a stair.
    :type paras: list Autodesk.Revit.DB.BuiltInParameterDefinition
    :return: List of element ids representing stair types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    stair_type = doc.GetElement(stair_type_id)
    for p_def in paras:
        p_value = rParaGet.get_built_in_parameter_value(stair_type, p_def)
        if p_value != None and p_value not in ids:
            ids.append(p_value)
    return ids


def get_all_similar_type_ids(doc, ids):
    """
    Gets all unique ids of similar types of element ids passed in.
    TODO: check for similar function elsewhere!
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param ids: list of type ids to be added to.
    :type ids: list of Autodesk.Revit.ElementIds
    :return: List of unique ids of similar types.
    :rtype: list of Autodesk.Revit.ElementIds
    """

    sim_ids = []
    for id in ids:
        el = doc.GetElement(id)
        sim_types = el.GetSimilarTypes()
        for st in sim_types:
            if st not in sim_ids:
                sim_ids.append(st)
    return sim_ids


def build_system_family_dictionary(doc, ids):
    """
    Returns dictionary where key is the system family name and values list of available type ids of that system family.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param ids: List of system family ids.
    :type ids: list of Autodesk.Revit.ElementIds
    :return: Dictionary of unique ids of similar types.
    :rtype: dictionary {str: list of Autodesk.Revit.ElementIds }
    """

    dic = {}
    for id in ids:
        el = doc.GetElement(id)
        if el.FamilyName in dic:
            dic[el.FamilyName].append(id)
        else:
            dic[el.FamilyName] = [id]
    return dic


def check_system_families(doc, ids, leave_one_behind):
    """
    Check whether a list of ids of system family is the entire list of types available in the model. If so it will remove one\
    type id per system family to allow safe purging.
    Revit requires at least one type definition per system family to be in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param ids: List of ids to check
    :type ids: list of Autodesk.Revit.ElementIds
    :param leave_one_behind: True: at least one type will be omitted from list.
    :type leave_one_behind: bool
    :return: List of unique ids of similar types.
    :rtype: list of Autodesk.Revit.ElementIds
    """

    dic_to_check = build_system_family_dictionary(doc, ids)
    similar_ids = get_all_similar_type_ids(doc, ids)
    dic_reference = build_system_family_dictionary(doc, similar_ids)
    ids = []
    for key, value in dic_to_check.items():
        if key in dic_reference:
            if len(dic_reference[key]) == len(dic_to_check[key]) and leave_one_behind:
                # need to leave one behind...
                if len(dic_to_check[key]) > 0:
                    dic_to_check[key].pop(0)
                    ids = ids + dic_to_check[key]
            else:
                ids = ids + dic_to_check[key]
        else:
            ids = ids + dic_to_check[key]
    return ids


def get_used_sub_types(doc, available_ids_getter, paras, leave_one_behind=True):
    """
    Returns a list of type ids which are not used in any stair types.
    Type ids are provided via an id getter function
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param available_ids_getter: function returning available type ids
    :type available_ids_getter: func(doc)
    :param paras: list of built in parameters attached to a stair type for given sub types (stringers, path, run, landing)
    :type paras: list of Autodesk.Revit.DB.BuiltInParameter
    :param leave_one_behind: _description_, defaults to True
    :type leave_one_behind: bool, optional
    :return: List of element ids
    :rtype: list of Autodesk.Revit.ElementIds
    """

    ids = []
    # get all available type ids and then check against all Stair type ids
    ids_available = available_ids_getter(doc)
    all_used_stair_type_ids = rStair.get_all_stair_type_ids_by_category(doc)
    ids_used_types = []
    for used in all_used_stair_type_ids:
        ids_used = get_used_sub_type_ids_from_stair_type(doc, used, paras)
        for id in ids_used:
            if id not in ids_used_types:
                ids_used_types.append(id)
    for id_available in ids_available:
        if id_available not in ids_used_types:
            ids.append(id_available)
    # need to check that we are not trying to delete last type of a system family....
    ids = check_system_families(doc, ids, leave_one_behind)
    return ids


# --------------------------------- purging subtypes ------------------------------------------------


def get_unused_stair_path_type_ids_to_purge(doc):
    """
    Gets all unused Stair path ids to purge, will omit on path type id per system family if none are used.
    This method can be used to safely delete unused stair path types. In the case that no stair\
        path instance using any of the types is placed, this will return all but one type id since\
        Revit requires at least one stair path type definition to be in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids 
    :rtype: list of Autodesk.Revit.ElementIds
    """

    ids_used = []
    available_types = rStairPath.get_stair_path_types_ids_by_class(doc)
    col = rStairPath.get_all_stair_path_instances(doc)
    for c in col:
        if c.GetTypeId() not in ids_used:
            ids_used.append(c.GetTypeId())
    ids = []
    for at in available_types:
        if at not in ids_used:
            ids.append(at)
    ids = check_system_families(doc, ids, True)
    return ids


def get_unused_stair_landing_type_ids_to_purge(doc):
    """
    Gets all unused Stair landing type ids.
    This method can be used to safely delete unused stair landing types. In the case that no stair\
        landing instance using any of the types is placed, this will return all but one type id since\
        Revit requires at least one stair landing type definition to be in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids 
    :rtype: list of Autodesk.Revit.ElementIds
    """

    ids = get_used_sub_types(
        doc,
        rStairLanding.get_stair_landing_types_ids_by_class,
        rStairLanding.STAIR_LANDING_TYPE_PARAS,
    )
    return ids


def get_unused_stair_run_type_ids_to_purge(doc):
    """
    Gets all unused Stair run type ids.
    This method can be used to safely delete unused stair run types. In the case that no stair\
        run instance using any of the types is placed, this will return all but one type id since\
        Revit requires at least one stair run type definition to be in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids 
    :rtype: list of Autodesk.Revit.ElementIds
    """

    ids = get_used_sub_types(
        doc, rStairRun.get_stair_run_types_ids_by_class, rStairRun.STAIR_RUN_TYPE_PARAS
    )
    return ids


def get_unused_stair_cut_mark_type_ids_to_purge(doc):
    """
    Gets all unused Stair cut mark type ids.
    This method can be used to safely delete unused stair cut mark types. In the case that no stair\
        cut mark instance using any of the types is placed, this will return all but one type id since\
        Revit requires at least one stair cut mark type definition to be in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids 
    :rtype: list of Autodesk.Revit.ElementIds
    """

    ids = get_used_sub_types(
        doc,
        rStairCut.get_stair_cut_mark_types_ids_by_class,
        rStairCut.STAIR_CUT_MARK_TYPE_PARAS,
    )
    return ids


def get_unused_stair_stringers_carriage_type_ids_to_purge(doc):
    """
    Gets all unused Stair stringer / carriage type ids.
    This method can be used to safely delete unused stair stringer / carriage types. In the case that no stair\
        string carriage instance using any of the types is placed, this will return all but one type id since\
        Revit requires at least one stringer carriage type definition to be in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids 
    :rtype: list of Autodesk.Revit.ElementIds
    """

    ids = get_used_sub_types(
        doc,
        rStairStringersAndCarriages.get_all_stair_stringers_carriage_type_ids_by_category,
        rStairStringersAndCarriages.STAIR_SUPPORT_TYPE_PARAS,
    )
    return ids


# -------------------------------- In place Stair types -------------------------------------------------------


def get_used_in_place_stair_type_ids(doc):
    """
    Gets all used in place stair type ids.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids
    :rtype: list of Autodesk.Revit.ElementIds
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, rStair.get_all_in_place_stair_type_ids, 1
    )
    return ids


def get_unused_in_place_stair_type_ids(doc):
    """
    Gets all unused in place stair type ids.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids
    :rtype: list of Autodesk.Revit.ElementIds
    """

    ids = rPurgeUtils.get_used_unused_type_ids(
        doc, rStair.get_all_in_place_stair_type_ids, 0
    )
    return ids


def get_unused_in_place_stair_type_ids_for_purge(doc):
    """
    Gets symbol (type) ids and family ids (when no type is in use) of in place Stair families which can be purged.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids
    :rtype: list of Autodesk.Revit.ElementIds
    """

    ids = rFamPurge.get_unused_in_place_ids_for_purge(
        doc, get_unused_in_place_stair_type_ids
    )
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit room creation. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Revit.Common.transaction import in_transaction

from duHast.Utilities.Objects.result import Result


from Autodesk.Revit.DB import Transaction

def create_room (doc, level, location_point, modify_action, transaction_manager=in_transaction):
    """
    Create a room in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param level: The level to create the room on.
    :type level: Autodesk.Revit.DB.Level
    :param location_point: The location point of the room.
    :type location_point: Autodesk.Revit.DB.XYZ
    :param  modify_action: An action executed on the room at the time of creation ( in same transaction).
    :type  modify_action: function taking the room as input
    :param transaction_manager: The transaction manager to use. Defaults to in_transaction.
    :type transaction_manager: TransactionManager, optional
    :return:
        Result class instance.

        - result.status. True if room was created successfully, otherwise  False.
        - Result.message. Generic success message.
        - Result.result. The created room if successful as first entry in list, otherwise None.

        On exception:

        - result.status (bool) will be False.
        - Result.message will contain the exception message.
        - Result.result will be an empty list.

    :rtype: Result
    """

    return_value = Result()

    try:
        
        def action():
            """
            Create the room and execute the modify action.
            """
            action_return_value = Result()
            try:
                # Create the room
                room = doc.Create.NewRoom(level, location_point)
                # Execute the modify action if provided
                if modify_action:
                    modify_action(room)
                action_return_value.result.append(room)
                action_return_value.append_message ("Room created successfully.")
            except Exception as e:
                action_return_value.update_sep (False,"{}".format(e))
            return action_return_value
        
        # execute inside transaction if a transaction manager is provided, otherwise run the action directly
        # assuming there is a transaction running already
        if transaction_manager:
            # Start the transaction
            tranny = Transaction(doc, "Create Room")
            return_value = transaction_manager(tranny, action)
        else:
            return_value = action()
        
        return return_value

    except Exception as e:
        return_value.update_sep (False,"{}".format(e))
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of utilities for an elements spatial properties
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import BuiltInParameter


def get_room_element_is_in_by_element_phase_created(element, phase_dict):
    """
    Get the room an element is in.The room phase need to match the phase the element is created in.

    :param element: The element to check
    :type element: FamilyInstance
    :param phase_dict: The dictionary of phase name to phase element id ( key is the phase name, value is the phase element id)
    :type phase_dict: dict

    :return: The room the element is in or None if no room exists or the element is not in a room
    :rtype: Room
    """

    # get the phase the element is created in
    el_phase = element.get_Parameter(BuiltInParameter.PHASE_CREATED)
    if el_phase != None:
        # get the phase name
        el_phase_name = el_phase.AsValueString()
        try:
            # get the phase element
            el_phase_elem = phase_dict[el_phase_name]
        except:
            print(
                "Element has phase created but phase not found: {}".format(
                    element.Id.IntegerValue
                )
            )
            return None
        # attempt to get the room the element is in
        # this can throw an exception if the element is created in a phase where there in no room enclosing the element
        try:
            el_room = element.get_Room(el_phase_elem)
            return el_room
        except:
            return None

    else:
        print("Element has no phase created: {}".format(element.Id.IntegerValue))
        return None



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit levels appearance modifier functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Note: Any level appearance modification in a view will throw an exception if the level is not actually visible in the view.

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import Autodesk.Revit.DB as rdb

from duHast.Utilities.Objects import result as res
from duHast.Revit.Common import transaction as rTran


def change_levels_2D(doc, levels, view):
    """
    Changes all levels in view to 2D

    Note: Any level past in, which is not visible in the view, will throw an exception when attempting to set to 2D.

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param levels: List of levels to be changed to 2D.
    :type levels: [Autodesk.Revit.DB.Level]
    :param view: The view in which to change the levels
    :type view: Autodesk.Revit.DB.View
    :return:
        Result class instance.
        - result.status. True if all levels where set to 2D, otherwise False.
        - result.message will contain the name(s) of the level(s) changed to 2D
        - result.result empty list
        On exception:
        - result.status (bool) will be False.
        - result.message will contain generic exception message including the level name.
        - result.result will be empty
    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    # needs to run in a transaction
    def action():
        action_return_value = res.Result()
        level_counter = 0
        for g in levels:
            level_counter = level_counter + 1
            try:
                g.SetDatumExtentType(
                    rdb.DatumEnds.End1, view, rdb.DatumExtentType.ViewSpecific
                )
                g.SetDatumExtentType(
                    rdb.DatumEnds.End0, view, rdb.DatumExtentType.ViewSpecific
                )
                action_return_value.update_sep(
                    True, "Changed level {} to 2D.".format(g.Name)
                )
            except Exception as e:
                action_return_value.update_sep(
                    False,
                    "Failed to change level {} to 2D with exception: {}".format(
                        g.Name, e
                    ),
                )
        if level_counter == 0:
            action_return_value.update_sep(
                True, "No levels visible in view {}".format(view.Name)
            )
        return action_return_value

    transaction = rdb.Transaction(doc, "levels to 2D")
    return_value = rTran.in_transaction(transaction, action)
    return return_value


def show_head_end(doc, level, view, end_identifier, show_head):
    """
    Toggles level head visibility on specified end for given level.

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param level: The level of which a heads visibility is to be toggled.
    :type level: Autodesk.Revit.DB.Level
    :param view: The view in which a level heads visibility is to be toggled.
    :type view: Autodesk.Revit.DB.View
    :param end_identifier: which end.
    :type end_identifier:
    :param show_head: True head will switched on, False it will be switched off
    :type show_head: bool
    :return:
        Result class instance.
        - result.status. True if all levels head(s) visibility was set successfully, otherwise False.
        - result.message will contain the name(s) of the level(s) where a head visibility was set.
        - result.result empty list
        On exception:
        - result.status (bool) will be False.
        - result.message will contain generic exception message including the level name.
        - result.result will be empty
    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    # needs to run in a transaction
    def action():
        action_return_value = res.Result()
        try:
            if show_head:
                level.ShowBubbleInView(end_identifier, view)
                action_return_value.update_sep(
                    True,
                    "Set level {} head to visible at end: {}".format(
                        level.Name, end_identifier
                    ),
                )
            else:
                level.HideBubbleInView(end_identifier, view)
                action_return_value.update_sep(
                    True,
                    "Set level {} head to invisible at end: {}".format(
                        level.Name, end_identifier
                    ),
                )
        except Exception as e:
            action_return_value.update_sep(
                False,
                "Failed to change level {} head visibility at end {} with exception: {}".format(
                    level.Name, end_identifier, e
                ),
            )
        return action_return_value

    transaction = rdb.Transaction(doc, "Toggle head. {}".format(show_head))
    return_value = rTran.in_transaction(transaction, action)
    return return_value


def hide_both_heads(doc, levels, view):
    """
    Hides both heads of levels in the given view.

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document

    :param level: The level of which a head's visibility is to be toggled.
    :type level: Autodesk.Revit.DB.Level

    :param view: The view in which a level head's visibility is to be toggled.
    :type view: Autodesk.Revit.DB.View

    :return: Result class instance.

        - `result.status` (bool): True if all levels' head(s) visibility was switched off successfully, otherwise False.
        - `result.message` (str): Contains the name(s) of the level(s) where a head visibility was set.
        - `result.result` (list): Empty list.

    On exception:

        - `result.status` (bool) will be False.
        - `result.message` (str) will contain a generic exception message including the level name.
        - `result.result` (list) will be empty.

    :rtype: :class:`.Result`"""

    return_value = res.Result()
    for l in levels:
        return_value.update(show_head_end(doc, l, view, rdb.DatumEnds.End1, False))
        return_value.update(show_head_end(doc, l, view, rdb.DatumEnds.End0, False))

    return return_value


def show_head_zero_end(doc, levels, view):
    """
    Turns on level heads at zero end in specified view.

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param level: The levels of which a heads visibility is to be toggled.
    :type level: [Autodesk.Revit.DB.Level]
    :param view: The view in which a level heads visibility is to be toggled.
    :type view: Autodesk.Revit.DB.View
    :return: Result class instance.

        - result.status. True if all levels head(s) visibility at zero end was set to visible successfully, otherwise False.
        - result.message will contain the name(s) of the level(s) where a head visibility was set.
        - result.result empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain generic exception message including the level name.
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    for l in levels:
        return_value.update(show_head_end(doc, l, view, rdb.DatumEnds.End0, True))

    return return_value


def show_head_one_end(doc, levels, view):
    """
    Turns on level heads at One end in specified view.

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param level: The levels of which a heads visibility is to be toggled.
    :type level: [Autodesk.Revit.DB.Level]
    :param view: The view in which a level heads visibility is to be toggled.
    :type view: Autodesk.Revit.DB.View
    :return: Result class instance.

        - result.status. True if all levels head(s) visibility at one end was set to visible successfully, otherwise False.
        - result.message will contain the name(s) of the level(s) where a head visibility was set.
        - result.result empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain generic exception message including the level name.
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    for l in levels:
        return_value.update(show_head_end(doc, l, view, rdb.DatumEnds.End1, True))

    return return_value


def toggle_head_end(doc, level, view, end_identifier):
    """
    Toggles level head visibility on specified end for given level in given views.

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param level: The level of which a heads visibility is to be toggled.
    :type level: Autodesk.Revit.DB.Level
    :param view: The view in which a level heads visibility is to be toggled.
    :type view: Autodesk.Revit.DB.View
    :param end_identifier: The end of the level to be modified.
    :type view: Autodesk.Revit.DB.DatumEnds
    :return: Result class instance.

        - result.status. True if all levels head(s) visibility was changed successfully, otherwise False.
        - result.message will contain the name(s) of the level(s) where a head visibility was changed.
        - result.result empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain generic exception message including the level name.
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    def action():
        try:
            action_return_value = res.Result()
            end_head_one = level.IsBubbleVisibleInView(end_identifier, view)
            if end_head_one == False:
                level.ShowBubbleInView(end_identifier, view)
                action_return_value.update_sep(
                    True,
                    "Set level {} head to visible at end: {}.".format(
                        level.Name, end_identifier
                    ),
                )
            else:
                level.HideBubbleInView(end_identifier, view)
                action_return_value.update_sep(
                    True,
                    "Set level {} head to not visible at end: {}.".format(
                        level.Name, end_identifier
                    ),
                )
        except Exception as e:
            action_return_value.update_sep(
                False,
                "Failed to change level {} head visibility at end: {} with exception: {}".format(
                    level.Name, end_identifier, e
                ),
            )
        return action_return_value

    transaction = rdb.Transaction(doc, "Toggle head.")
    return_value = rTran.in_transaction(transaction, action)
    return return_value


def toggle_head_one_end(doc, levels, view):
    """
    Toggles level head visibility on one end for given levels

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param levels: The levels of which a heads visibility at one end is to be toggled.
    :type levels: [Autodesk.Revit.DB.Level]
    :param view: The view in which a level heads visibility is to be toggled.
    :type view: Autodesk.Revit.DB.View
    :return: Result class instance.

        - result.status. True if all levels head(s) visibility at one end was changed successfully, otherwise False.
        - result.message will contain the name(s) of the level(s) where a head visibility was changed.
        - result.result empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain generic exception message including the level name.
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    for l in levels:
        return_value.update(toggle_head_end(doc, l, view, rdb.DatumEnds.End1))
    return return_value


def toggle_head_zero_end(doc, levels, view):
    """
    Toggles level head visibility on zero end for given levels

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param levels: The levels of which a head visibility at zero end is to be toggled.
    :type levels: [Autodesk.Revit.DB.Level]
    :param view: The view in which a level heads visibility is to be toggled.
    :type view: Autodesk.Revit.DB.View
    :return: Result class instance.

        - result.status. True if all level head(s) visibility at one end was changed successfully, otherwise False.
        - result.message will contain the name(s) of the level(s) where a head visibility was changed.
        - result.result empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain generic exception message including the level name.
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    for l in levels:
        return_value.update(toggle_head_end(doc, l, view, rdb.DatumEnds.End0))
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of functions around Revit filled regions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.result import Result

from duHast.Revit.DetailItems.Utility import detail_items_type_sorting as rDetailItemTypeSort
from duHast.Revit.DetailItems.detail_items import get_all_detail_types_by_category, FILLED_REGION_TYPE
from duHast.Revit.Common.parameter_get_utils import get_built_in_parameter_value, getter_double_as_double_converted_to_metric
from duHast.Revit.Common.Geometry.curve_loops import get_area_from_closed_curve_loop

# import Autodesk
from Autodesk.Revit.DB import (
    BuiltInParameter,
    FilledRegion,
    FilteredElementCollector,
)


def get_filled_regions_in_model(doc):
    """
    Gets all filled region instances in a model.

    Filters by class.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A list containing floor instances.
    :rtype: list Autodesk.Revit.DB.FilledRegion
    """

    return FilteredElementCollector(doc).OfClass(FilledRegion)


def get_all_filled_region_type_ids_available(doc):
    """
    Gets all filled region types ids in model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A list of element ids representing filled region types.
    :rtype: list Autodesk.Revit.DB.ElementIds
    """

    dic = rDetailItemTypeSort.build_detail_type_ids_dictionary(
        get_all_detail_types_by_category(doc)
    )
    if FILLED_REGION_TYPE in dic:
        return dic[FILLED_REGION_TYPE]
    else:
        return []



# --------------------------------------------- geometry ------------------

def get_filled_region_curve_loops(filled_region):
    """
    Gets the curve loops of a filled region.
    :param filled_region: A filled region instance.
    :type filled_region: Autodesk.Revit.DB.FilledRegion

    :return: A list of curve loops.
    :rtype: list Autodesk.Revit.DB.CurveLoop
    """

    # get the loops of a filled region's geometry
    curve_loops = filled_region.GetBoundaries()

    return curve_loops


def get_filled_region_loop_area(doc, view, filled_region, loop_index=0):
    """
    Gets loop area of a filled region.
    This will create a temporary filled region using just the loop and get the area.

    :param filled_region: A filled region instance.
    :type filled_region: Autodesk.Revit.DB.FilledRegion
    :param loop_index: The index of the loop to get the area for.
    :type loop_index: int
    
    :return: The area of the loop.
    :rtype: float
    """
    
    # set up a status tracker
    return_value = Result()
    
    try:
        # get the loops of a filled region's geometry
        curve_loops = get_filled_region_curve_loops(filled_region)
        
        if loop_index >= len(curve_loops):
            return_value.update_sep(False, "Loop index out of range.")
            return return_value

        # get the area of the filled region using the loop
        area_result = get_area_from_closed_curve_loop(
            doc=doc,
            view=view,
            curve_loop=curve_loops[loop_index],
        )

        # update the return value with the area
        return_value.update(area_result)
    except Exception as e:
        return_value.update_sep(False, "Failed to get filled region loop area with error: {}".format(e))
    
    return return_value

def get_filled_region_area(filled_region):
    """
    Gets the area of a filled region using the built-in parameter.
   
    :param filled_region: A filled region instance.
    :type filled_region: Autodesk.Revit.DB.FilledRegion

    :return: The area of the filled region.
    :rtype: float
    """

    # set up a status tracker
    return_value = Result()

    try:
       
        # get the area of the filled region
        area = get_built_in_parameter_value(
            element=filled_region,
            built_in_parameter_def=BuiltInParameter.HOST_AREA_COMPUTED,
            parameter_value_getter=getter_double_as_double_converted_to_metric,
        )
        
        # update the return value with the area
        return_value.result.append(area)
        return_value.append_message("Filled region area retrieved successfully.")
    except Exception as e:
        return_value.update_sep(False, "Failed to get filled region area with error: {}".format(e))

    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A class to provide filtering of views of observable collections..
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

IsChecked property should be bound similar to this sample code:
<GridViewColumn.HeaderContainerStyle>
    <Style TargetType="GridViewColumnHeader">
        <Setter Property="ContextMenu">
            <Setter.Value>
                <ContextMenu>
                    <ItemsControl ItemsSource="{Binding UniqueFamilyNames}">
                        <ItemsControl.ItemTemplate>
                            <DataTemplate>
                                <CheckBox 
                                    Content="{Binding Value}" 
                                    IsChecked="{Binding IsChecked, Mode=TwoWay}"   <<<--- here
                                    Command="{Binding CheckCommand}"/>
                            </DataTemplate>
                        </ItemsControl.ItemTemplate>
                    </ItemsControl>
                </ContextMenu>
            </Setter.Value>
        </Setter>
    </Style>
    </GridViewColumn.HeaderContainerStyle>

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.UI.Objects.WPF.ViewModels.ViewModelBase import ViewModelBase

class FilterItem(ViewModelBase):
    def __init__(self, value, refresh_view_method):
        """
        A filter item for a view model view. 
        
        Is used in column context menu filters to filter the view based on the value of the item.
        
        :param value: the value of the item
        :param refresh_view_method: the method to call to refresh the view
        """
        
        super(FilterItem, self).__init__()
        self._value = value
        self.refresh_view_method = refresh_view_method
        
        # default to checked
        self._is_checked = True

    @property
    def Value(self):
        """
        Gets the value of the item displayed in the context menu.
        """
        
        return self._value
    
    @property
    def IsChecked(self):
        """
        Gets the is checked value of the item.
        """
        
        return self._is_checked

    @IsChecked.setter
    def IsChecked(self, value):
        """
        Sets the is checked value of the item.
        """
        
        if  self._is_checked != value:
            self._is_checked = value
            self.RaisePropertyChanged("IsChecked")
            
            # refresh the view to show the changes
            if self.refresh_view_method:
                self.refresh_view_method()

        




"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit grids extend to view crop.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr

from Autodesk.Revit.DB import (
    CurveLoop,
    DatumExtentType,
    IntersectionResultArray,
    Line,
    SetComparisonResult,
    Transaction,
    Transform,
    ViewType,
    XYZ,
)

from duHast.Revit.Common import transaction as rTran
from duHast.Revit.Views import views as rView
from duHast.Utilities.Objects import result as res
from duHast.Revit.Grids.grids import get_linear_grids_in_model_by_view
from duHast.Revit.Common.Geometry.points import (
    on_which_side_of_line_is_point,
    distance_between_two_points,
)
from duHast.Revit.Common.Geometry.curve import get_perpendicular_line_through_point


""" type of views this script will work with"""
TEMPLATE_VIEW_TYPES = [ViewType.FloorPlan, ViewType.CeilingPlan, ViewType.AreaPlan]


def _check_active_view_type(view):
    """
    Checks whether the past in view is of a specific type.

    Refer global list: TEMPLATE_VIEW_TYPES

    :param view: The view.
    :type view: AutoDesk.Revit.DB.View
    :return: True if view is of a given view type, otherwise False
    :rtype: bool
    """

    if view.ViewType in TEMPLATE_VIEW_TYPES:
        return True
    else:
        return False


def _sort_projected_grid_end_points(grid, rotation_axis, projected_points):
    """
    Sorts projected grid end points.

    Checks on which side of a line perpendicular through the grid origin the projected points and the other grid end point are
    Possible outcomes:

    - all three points are smaller then 0 or all three points are greater then 0: grid 0 point is projected point closest to grid origin
    - one projected point is smaller then 0 and the other one is greater then 0: grid 0 point is point on opposite site to other grid end point
    - both projected points are on the opposite site then other grid origin point: grid 0 point is point furthest from grid origin

    :param grid: The grid to extend to the view crop.
    :type grid: Autodesk.Revit.DB.Grid
    :param rotation_axis: The axis around which to rotate the test line.
    :type rotation_axis: Autodesk.Revit.DB.XYZ
    :param projected_points: Two projected points representing the new grid extent.
    :type projected_points: [Autodesk.Revit.DB.XYZ,Autodesk.Revit.DB.XYZ]

    :return: Sorted list of points where point at index 0 represents the start point of the new grid extent and point at index 1 the end point of the new grid.
    :rtype: [Autodesk.Revit.DB.XYZ,Autodesk.Revit.DB.XYZ]
    """

    # list of points to be returned
    sorted_points = []
    # get grid details required for rotation
    curve_grid = grid.Curve
    # origin point is the same as 0 end?
    grid_origin_point = curve_grid.Origin
    # create line through grid origin rotated by 90 degree
    reference_line = get_perpendicular_line_through_point(
        curve_grid, rotation_axis, grid_origin_point
    )

    # work out on which side of the reference line are the two projected points and the other grid end point
    grid_end_point_side = on_which_side_of_line_is_point(
        reference_line, curve_grid.GetEndPoint(1)
    )
    projected_grid_end_point_zero = on_which_side_of_line_is_point(
        reference_line, projected_points[0]
    )
    projected_grid_end_point_one = on_which_side_of_line_is_point(
        reference_line, projected_points[1]
    )

    # work out distances from grid origin to projected points
    projected_end_point_zero_to_grid_origin = distance_between_two_points(
        grid_origin_point, projected_points[0]
    )
    projected_end_point_one_to_grid_origin = distance_between_two_points(
        grid_origin_point, projected_points[1]
    )

    # possible outcomes
    # all three points are smaller then 0 or all three points are greater then 0: grid 0 point is projected point closest to grid origin
    # one projected point is smaller then 0 and the other one is greater then 0: grid 0 point is point on opposite site to other grid end point
    # both projected points are on the opposite site then other grid origin point: grid 0 point is point furthest from grid origin

    if (
        grid_end_point_side < 0
        and projected_grid_end_point_zero < 0
        and projected_grid_end_point_one < 0
    ) or (
        grid_end_point_side > 0
        and projected_grid_end_point_zero > 0
        and projected_grid_end_point_one > 0
    ):
        # grid 0 point is projected point closest to grid origin
        if (
            projected_end_point_zero_to_grid_origin
            < projected_end_point_one_to_grid_origin
        ):
            sorted_points.append(projected_points[0])
            sorted_points.append(projected_points[1])
        else:
            sorted_points.append(projected_points[1])
            sorted_points.append(projected_points[0])
    elif projected_grid_end_point_zero * projected_grid_end_point_one < 0:
        # grid 0 point is point on opposite site to other grid end point
        if projected_grid_end_point_zero * grid_end_point_side < 0:
            # projected 0 end point is opposite grid origin point
            sorted_points.append(projected_points[0])
            sorted_points.append(projected_points[1])
        else:
            # projected 1 end point is opposite grid origin point
            sorted_points.append(projected_points[1])
            sorted_points.append(projected_points[0])
    else:
        # grid 0 point is point furthest from grid origin
        if (
            projected_end_point_zero_to_grid_origin
            > projected_end_point_one_to_grid_origin
        ):
            sorted_points.append(projected_points[0])
            sorted_points.append(projected_points[1])
        else:
            sorted_points.append(projected_points[1])
            sorted_points.append(projected_points[0])
    return sorted_points


def _create_datum_line_linear_grid(crop_box_boundary_lines, grid, debug_view):
    """
        Creates a new datum line for a linear grid from its intersection points with a crop box.

        :param crop_box_boundary_lines: The boundary segments of a view crop.
        :type crop_box_boundary_lines: _type_
        :param grid: The grid to be extended.
        :type grid: Autodesk.Revit.DB.Grid
        :param debug_view: The view to draw debug lines on. (commands commented out in the moment)
        :type debug_view: Autodesk.Revit.DB.View

        :return:
        Result class instance.

        - result.status. True if a new datum line was successfully calculated, otherwise False.
        - result.message will contain the name(s) of the grids.
        - result.result empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain generic exception message.
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    new_grid_line = None
    # get the curve describing the grid (Autodesk.Revit.DB.Line)
    curve_grid = grid.Curve
    # check this is a line (and not an arc or similar...)
    if type(curve_grid) == Line:
        # get a direction of that curve
        grid_vector = curve_grid.Direction
        projected_points_list = []
        # loop over view crop outlines
        for line_bound in crop_box_boundary_lines:
            # cast a ray from the grid origin point along the direction vector
            point = XYZ(
                curve_grid.Origin.X, curve_grid.Origin.Y, line_bound.GetEndPoint(0).Z
            )
            ray_cast = Line.CreateUnbound(point, grid_vector)

            # set up result array that will contain the intersection point if one is found
            out_intersection_results = clr.Reference[IntersectionResultArray]()
            # perform a ray intersect operation of the unbound line through the grid curve origin and the bounding box line and populate IntersectionResultArray if there is an intersection
            result = ray_cast.Intersect(line_bound, out_intersection_results)
            # check what type of intersection the ray has with the bounding box line
            # if result is an overlap store the intersection point
            if result == SetComparisonResult.Overlap:
                intersection_result = out_intersection_results.Value
                projected_points_list.append(intersection_result[0].XYZPoint)

        # check if we have 2 intersection points
        if len(projected_points_list) == 2:
            # create a new grid line curve
            # Check location of points...
            # use approach checking which new point is on which side of the original origin point if drawing a perpendicular line through the curve origin point:
            # if both new points are on the same side than the other original curve end point then the point closest to the original origin is the new end 0 and the other one is the new end 1
            # if both new points are on the opposite side than the other original curve end point then the point closest to the original origin is the new end 0 and the other one is the new end 1
            # if one new point is on the opposing side and one new point on the same side than the other original curve end point, then the point on the opposing side is the new end 0 and the other one is the new end 1
            sorted_points = _sort_projected_grid_end_points(
                grid, XYZ.BasisZ, projected_points_list
            )

            # for some reason the create bound needs to get the points back to front in order not to swap end 0 and end 1 once the curve is assigned.
            new_grid_line = Line.CreateBound(sorted_points[1], sorted_points[0])
            return_value.update_sep(
                True, "Grid {} line successfully calculated".format(grid.Name)
            )
            return_value.result = [new_grid_line]
        else:
            return_value.update_sep(
                False,
                "Grid {} returned {} intersection points. Exactly 2 intersection points are required.".format(
                    grid.Name, len(projected_points_list)
                ),
            )
    else:
        return_value.update_sep(
            False, "Grid {} is not line based and will be ignored.".format(grid.Name)
        )
    return return_value


def _extend_linear_grids_to_view_crop(
    active_view, bound_lines_original, cut_offset, grids
):
    """
    Extends grids to view crop.

    :param active_view: The view in which the grids to extend to the crop.
    :type active_view: Autodesk.Revit.DB.View
    :param bound_lines_original: List of lines forming the view crop.
    :type bound_lines_original: IList<AutoDesk.Revit.DB.CurveLoop>
    :param cut_offset: Z value of view.
    :type cut_offset: double
    :param grids: Linear grids visible in view.
    :type grids: [Autodesk.Revit.DB.Grid]

    :return:
    Result class instance.

    - result.status. True if all grids in view got extended successfully, otherwise False.
    - result.message will contain the name(s) of the grids extended.
    - result.result empty list

    On exception:

    - result.status (bool) will be False.
    - result.message will contain generic exception message.
    - result.result will be empty

    :rtype: :class:`.Result`
    """
    return_value = res.Result()
    try:
        # check if plan view...crop box needs a transform!
        if active_view.ViewDirection.IsAlmostEqualTo(XYZ(0, 0, 1)):
            # get Current Elevation of boundLines
            current_view_plane_z = list(bound_lines_original)[0].GetEndPoint(0).Z
            # transform boundLines CurveLoop
            crop_box_transform = Transform.CreateTranslation(
                XYZ(0, 0, cut_offset - current_view_plane_z)
            )
            # transform crop box curve loop
            bound_lines_transformed = CurveLoop.CreateViaTransform(
                bound_lines_original, crop_box_transform
            )
            # loop over grids and calculate new curves intersecting the crop box curve loop
            for grid in grids:
                # calculate grid extent to fit exactly into the crop box
                new_grid_line_curve_status = _create_datum_line_linear_grid(
                    bound_lines_transformed, grid, active_view
                )
                return_value.update(new_grid_line_curve_status)
                if new_grid_line_curve_status.status:
                    new_grid_line_curve = new_grid_line_curve_status.result[0]
                    try:
                        # set grid to be a 2D grid, in active view with curve as calculated.
                        grid.SetCurveInView(
                            DatumExtentType.ViewSpecific,
                            active_view,
                            new_grid_line_curve,
                        )
                        return_value.update_sep(
                            True,
                            "Set extend of grid: {} in view: {}".format(
                                grid.Name, active_view.Name
                            ),
                        )
                    except Exception as e:
                        return_value.update_sep(
                            False,
                            "Failed to set extend of grid: {} with: {}".format(
                                grid.Name, e
                            ),
                        )
        else:
            # elevations and sections
            for grid in grids:
                # calculate grid extent to fit exactly into the crop box
                new_grid_line_curve = _create_datum_line_linear_grid(
                    bound_lines_original, grid, active_view
                )
                if new_grid_line_curve is not None:
                    try:
                        # set grid to be a 2D grid, in active view with curve as calculated.
                        grid.SetCurveInView(
                            DatumExtentType.ViewSpecific,
                            active_view,
                            new_grid_line_curve,
                        )
                        return_value.update_sep(
                            True,
                            "Set extend of grid: {} in view: {}".format(
                                grid.Name, active_view.Name
                            ),
                        )
                    except Exception as e:
                        return_value.update_sep(
                            False,
                            "Failed to set extend of grid: {} with: {}".format(
                                grid.Name, e
                            ),
                        )
    except Exception as e:
        return_value.update_sep(False, e.message)
    return return_value


def extend_linear_grids_to_crop_box_of_view(doc, view):
    """
    Extends all linear grids to intersection points with crop box on active view.

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document

    :return:
    Result class instance.

    - result.status. True if all grids in active view got extended successfully, otherwise False.
    - result.message will contain the name(s) of the grids extended.
    - result.result empty list

    On exception:

    - result.status (bool) will be False.
    - result.message will contain generic exception message.
    - result.result will be empty

    :rtype: :class:`.Result`
    """
    return_value = res.Result()
    try:
        # check the active view type
        if not (_check_active_view_type(view)):
            raise ValueError(
                "The active view {} is not of an supported view type: {}.".format(
                    view.Name, view.ViewType
                )
            )
        # get grids visible in active view
        # note: this will only return standalone linear grids
        linear_grids_in_view = get_linear_grids_in_model_by_view(doc, view)
        # check if anything came back
        if len(linear_grids_in_view) == 0:
            raise ValueError(
                "The active view {} contains no grids.".format(
                    view.Name,
                )
            )
        # get Z point of active view
        cut_offset = (
            linear_grids_in_view[0]
            .GetCurvesInView(DatumExtentType.ViewSpecific, view)[0]
            .GetEndPoint(0)
            .Z
        )
        # get the crop box outline
        shape_manager = view.GetCropRegionShapeManager()
        # check whether crop box outline is valid (i.e. the view has crop active either just as visible or as visible and applied)
        bound_lines_original = None
        if len(shape_manager.GetCropShape()) > 0:
            bound_lines_original = shape_manager.GetCropShape()[0]
        else:
            raise ValueError(
                "The active view {} does has no crop active.".format(view.Name)
            )

        # change grids in view within an action (which will get wrapped in a transaction)
        def action():
            return_value = _extend_linear_grids_to_view_crop(
                view,  # the active view
                bound_lines_original,  # the crop bound lines
                cut_offset,  # z-point of active view
                linear_grids_in_view,  # grids in view
            )
            return return_value

        transaction = Transaction(
            doc, "Updating grid extents in view: {}".format(view.Name)
        )
        # execute the transaction
        updated_grids = rTran.in_transaction(transaction, action)
        return_value.update(updated_grids)
    except Exception as e:
        return_value.update_sep(False, e.message)
        return_value.append_message("Aborting!")
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit shared parameters.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System

# import common library modules
from duHast.Revit.Common.revit_version import get_revit_version_number
from duHast.Revit.SharedParameters.shared_parameter_add import (
    load_shared_parameter_file,
)

# import Autodesk
import Autodesk.Revit.DB as rdb

# --------------------------------------------- utility functions ------------------


def get_all_shared_parameters(doc):
    """
    Gets all shared parameters in a model

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing shared parameter elements
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = rdb.FilteredElementCollector(doc).OfClass(rdb.SharedParameterElement)
    return collector


def get_all_shared_parameters_from_file(rvt_doc, shared_param_file_path=None):
    """
    Gets all the shared parameter definitions from a shared parameter file. Can
    either be set manually or will be loaded from the default location which Revit
    is pointing at
    :param rvt_doc: Current Revit model document.
    :type rvt_doc: Autodesk.Revit.DB.Document
    :param shared_param_file_path: (Optional) The path to the shared parameter file.
    :type shared_param_file_path: str
    :return: A flat list of shared parameter definitions
    :rtype: list
    """
    def_file = load_shared_parameter_file(rvt_doc, shared_param_file_path)
    shared_params = []

    if def_file:
        sp_groups = def_file.Groups
        for sp_gp in sp_groups:
            for sp_def in sp_gp.Definitions:
                shared_params.append(sp_def)

        return shared_params
    else:
        return None


def get_family_shared_parameters(doc):
    """
    Gets all family parameters which are shared parameters.

    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :raises Exception: "Document is not a family document." when a non family document is past in.

    :return: A list of family parameters
    :rtype: [Autodesk.Revit.DB.FamilyParameter]
    """

    if doc.IsFamilyDocument:
        fam_manager = doc.FamilyManager
        shared_fam_paras = []
        for fam_para in fam_manager.GetParameters():
            try:
                # only shared parameters hav .GUID property...
                if str(fam_para.GUID) != "":
                    shared_fam_paras.append(fam_para)
            except Exception as e:
                pass
    else:
        raise Exception("Document is not a family document.")
    return shared_fam_paras


def get_family_parameters(doc):
    """
    Gets all family parameters.

    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :raises Exception: "Document is not a family document." when a non family document is past in.

    :return: A list of family parameters
    :rtype: [Autodesk.Revit.DB.FamilyParameter]
    """

    if doc.IsFamilyDocument:
        fam_manager = doc.FamilyManager
        shared_fam_paras = []
        for fam_para in fam_manager.GetParameters():
            shared_fam_paras.append(fam_para)
    else:
        raise Exception("Document is not a family document.")
    return shared_fam_paras


# ------------------------------------------------------- parameter utilities --------------------------------------------------------------------


def check_whether_shared_parameters_are_in_file(doc, parameter_gui_ds):
    """
    Filters the past in list of shared parameter GUIDs by using the shared parameters in the document.
        Only parameter in both will be returned.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param parameter_gui_ds: list of shared parameter GUIDs as string values
    :type parameter_gui_ds: list str

    :return: list of shared parameter GUIDs as string values
    :rtype: list str
    """

    filtered_guids = []
    paras = get_all_shared_parameters(doc)
    for p in paras:
        if p.GuidValue.ToString() in parameter_gui_ds:
            filtered_guids.append(p.GuidValue.ToString())
    return filtered_guids


def check_whether_shared_parameters_by_name_is_family_parameter(doc, parameter_name):
    """
    Checks, by name, whether a shared parameter exists as a family parameter in a family.

    param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param parameter_name: The name of the parameter.
    :type parameter_name: str

    :return: A family parameter if match was found, otherwise None
    :rtype: Autodesk.Revit.DB.FamilyParameter
    """

    para = None
    paras = get_family_parameters(doc)
    for fam_para in paras:
        if fam_para.Definition.Name == parameter_name:
            try:
                # only shared parameters have .GUID property...
                if str(fam_para.GUID) != "":
                    para = fam_para
                    break
            except Exception as e:
                pass
    return para


def is_shared_parameter_definition_used(doc, shared_para):
    """
    Tests if a shared parameter GUID is used by a family parameter.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param shared_para: A shared parameter
    :type shared_para: Autodesk.Revit.DB.SharedParameterElement

    :return: True is match is found, otherwise False
    :rtype: bool
    """

    fam_shared_paras = get_family_shared_parameters(doc)
    match = False
    for fam_shared_para in fam_shared_paras:
        if fam_shared_para.GUID == shared_para.GuidValue:
            match = True
            break
    return match


def get_unused_shared_parameter_definitions(doc):
    """
    Returns all unused shard parameter definitions in a family document.

    Note: These shared parameters might be used in any nested family!

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A list of shared parameters
    :rtype: [Autodesk.Revit.DB.SharedParameterElement]
    """

    fam_shared_paras = get_family_shared_parameters(doc)
    shared_paras = get_all_shared_parameters(doc)
    unused_shared_parameter_definition = []
    for shared_para in shared_paras:
        match = False
        for fam_shared_para in fam_shared_paras:
            if fam_shared_para.GUID == shared_para.GuidValue:
                match = True
                break
        if match == False:
            unused_shared_parameter_definition.append(shared_para)
    return unused_shared_parameter_definition


def get_shared_parameter_definition(parameter_name, def_file):
    """
    Returns a shared parameter definition from a shared parameter file.

    :param parameter_name: The shared parameter name.
    :type parameter_name: str
    :param def_file: The shared parameter file definition.
    :type def_file: Autodesk.Revit.DB.DefinitionFile

    :return: The shared parameter definition. None if no parameter with a matching name was found.
    :rtype: Autodesk.Revit.DB.ExternalDefinition
    """

    parameter_definition = None
    try:
        # loop through parameters and try to find matching one
        # loop through all definition groups
        for group in def_file.Groups:
            # loop through para's within definition group
            for def_para in group.Definitions:
                # check whether this is the parameter we are after
                if def_para.Name == parameter_name:
                    # match and out
                    parameter_definition = def_para
                    break
            if parameter_definition != None:
                break
    except Exception as e:
        pass
    return parameter_definition


# ------------------------------------------------------- parameter reporting --------------------------------------------------------------------


def param_binding_exists(doc, param_name, param_type):
    """
    Gets all parameter bindings for a given parameter depending on revit version.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param param_name: The name of the parameter.
    :type param_name: str
    :param param_type: The parameter type. (Area, vs text vs... (deprecated in Revit 2022!)
    :type param_type: Autodesk.Revit.DB.ParameterType

    :return: List of categories a parameter is attached to.
    :rtype: list of str
    """

    revit_version = get_revit_version_number(doc)
    data = []
    if revit_version <= 2022:
        data = param_binding_exists_2022(
            doc=doc, param_name=param_name, param_type=param_type
        )
    else:
        data = param_binding_exists_2023(
            doc=doc, param_name=param_name, type_id=param_type
        )
    return data


def param_binding_exists_2022(doc, param_name, param_type):
    """
    Gets all parameter bindings for a given parameter fro Revit versions up to 2022.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param param_name: The name of the parameter.
    :type param_name: str
    :param param_type: The parameter type. (Area, vs text vs... (deprecated in Revit 2022!)
    :type param_type: Autodesk.Revit.DB.ParameterType

    :return: List of categories a parameter is attached to.
    :rtype: list of str
    """

    categories = []
    map = doc.ParameterBindings
    iterator = map.ForwardIterator()
    iterator.Reset()
    while iterator.MoveNext():
        if (
            iterator.Key != None
            and iterator.Key.Name == param_name
            and iterator.Key.ParameterType == param_type
        ):
            elem_bind = iterator.Current
            for cat in elem_bind.Categories:
                categories.append(cat.Name)
            break
    return categories


def param_binding_exists_2023(doc, param_name, type_id):
    """
    Gets all parameter bindings for a given parameter for Revit 2023 onwards

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param param_name: The name of the parameter.
    :type param_name: str
    :param type_id: Forge type id
    :type type_id: Autodesk.Revit.DB.ForgeTypeId

    :return: List of categories a parameter is attached to.
    :rtype: list of str
    """

    categories = []
    map = doc.ParameterBindings
    iterator = map.ForwardIterator()
    iterator.Reset()
    while iterator.MoveNext():
        if (
            iterator.Key != None
            and iterator.Key.Name == param_name
            and iterator.Key.GetDataType() == type_id
        ):
            elem_bind = iterator.Current
            for cat in elem_bind.Categories:
                categories.append(cat.Name)
            break
    return categories



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit view sheet sets. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.result import Result
from duHast.Revit.Common.transaction import in_transaction

from Autodesk.Revit.DB import Transaction


from Autodesk.Revit.DB import Element, FilteredElementCollector, PrintRange, ViewSheet, ViewSheetSet, ViewSet


def get_view_sheet_sets(doc):
    """
    Get all view sets in the Revit document.
    :param doc: The Revit document to search for view sets.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing all view sets.
    :rtype: FilteredElementCollector
    """

    # set up a filtered element collector to get all view sets in the document
    view_sheet_set_collector = FilteredElementCollector(doc).OfClass(ViewSheetSet)
    return view_sheet_set_collector


def get_sheet_and_view_ids_from_view_sheet_set(view_sheet_set):
    """
    Extracts sheet and view IDs from a given view set.
    :param view_sheet_set: The view set from which to extract sheet and view IDs.
    :type view_sheet_set: Autodesk.Revit.DB.ViewSheetSet

    :return: A tuple containing two lists: sheet IDs and view IDs.
    :rtype: tuple
    """

    # Initialize lists to hold sheet and view IDs
    sheet_ids = []
    view_ids = []

    # Iterate through the ordered view list in the view set
    for element_in_set in view_sheet_set.OrderedViewList:

        # Check if the element is a ViewSheet
        if (isinstance(element_in_set, ViewSheet)):
            sheet_ids.append(element_in_set.Id)
        else: 
            view_ids.append(element_in_set.Id)
    
    return (sheet_ids, view_ids)


def get_sheets_and_views_from_view_sheet_set(view_sheet_set):
    """
    Extracts sheets and views from a given view set.
    :param view_sheet_set: The view set from which to extract sheet and view IDs.
    :type view_sheet_set: Autodesk.Revit.DB.ViewSheetSet

    :return: A tuple containing two lists: sheet IDs and view IDs.
    :rtype: tuple
    """

    # Initialize lists to hold sheet and view IDs
    sheets = []
    views = []

    # Iterate through the ordered view list in the view set
    for element_in_set in view_sheet_set.OrderedViewList:

        # Check if the element is a ViewSheet
        if (isinstance(element_in_set, ViewSheet)):
            sheets.append(element_in_set)
        else: 
            views.append(element_in_set)
    
    return (sheets, views)


def get_sheets_from_view_sheet_set(view_sheet_set):
    
    """
    Extracts only the sheets from a given view set.

    :param view_sheet_set: The view set from which to extract sheet objects.
    :type view_sheet_set: Autodesk.Revit.DB.ViewSheetSet

    :return: A list of sheet objects.
    :rtype: list
    """

    # Initialize lists to hold sheet and view IDs
    sheets, _ = get_sheets_and_views_from_view_sheet_set(view_sheet_set)

    # Return only the sheets
    return sheets


def get_views_from_view_sheet_set(view_sheet_set):
    
    """
    Extracts only the views from a given view set.

    :param view_sheet_set: The view set from which to extract sheet objects.
    :type view_sheet_set: Autodesk.Revit.DB.ViewSheetSet

    :return: A list of view objects.
    :rtype: list
    """

    # Initialize lists to hold sheet and view IDs
    _, views = get_sheets_and_views_from_view_sheet_set(view_sheet_set)

    # Return only the sheets
    return views


def get_sheet_ids_from_view_sheet_set(view_sheet_set):
    
    """
    Extracts only the sheet ids from a given view set.

    :param view_sheet_set: The view set from which to extract sheet objects.
    :type view_sheet_set: Autodesk.Revit.DB.ViewSheetSet

    :return: A list of Element Id objects.
    :rtype: list
    """

    # Initialize lists to hold sheet and view IDs
    sheet_ids, _ = get_sheet_and_view_ids_from_view_sheet_set(view_sheet_set)

    # Return only the sheets
    return sheet_ids


def get_view_ids_from_view_sheet_set(view_sheet_set):
    
    """
    Extracts only the view ids from a given view set.

    :param view_sheet_set: The view set from which to extract sheet objects.
    :type view_sheet_set: Autodesk.Revit.DB.ViewSheetSet

    :return: A list of Element Id objects.
    :rtype: list
    """

    # Initialize lists to hold sheet and view IDs
    _, view_ids = get_sheet_and_view_ids_from_view_sheet_set(view_sheet_set)

    # Return only the sheets
    return view_ids


def get_view_sheet_set_by_name(doc, view_sheet_set_name):
    """
    Retrieves a view set by its name from the Revit document.

    :param doc: The Revit document to search for the view set.
    :type doc: Autodesk.Revit.DB.Document
    :param view_sheet_set_name: The name of the view set to retrieve.
    :type view_sheet_set_name: str

    :return: The ViewSheetSet object if found, otherwise None.
    :rtype: Autodesk.Revit.DB.ViewSheetSet or None
    """

    # get all view sets in the document
    view_sheet_sets = get_view_sheet_sets(doc)

    # iterate through the view sets and return the one with the matching name
    for view_sheet_set in view_sheet_sets:
        if view_sheet_set.Name == view_sheet_set_name:
            return view_sheet_set
    
    return None


def get_current_view_sheet_settings_element(doc):
    """
    Retrieves the current ViewSheetSetting element from the Revit document.

    :param doc: The Revit document from which to retrieve the ViewSheetSetting.
    :type doc: Autodesk.Revit.DB.Document
    :return: The ViewSheetSetting element if found, otherwise None.
    :rtype: Autodesk.Revit.DB.ViewSheetSetting or None
    """

    # Access the PrintManager
    print_manager = doc.PrintManager

    # Set the PrintRange to Select - this is required!
    print_manager.PrintRange = PrintRange.Select

    # Now you can access the ViewSheetSetting
    view_sheet_setting = print_manager.ViewSheetSetting

    return view_sheet_setting


def update_view_sheet_set(doc, view_sheet_set, view_sheet_setting, sheets, views, clear_existing=True, transaction_manager = in_transaction, save_after_update=True):
    """
    Updates the view set with the given sheets and views.

    If either views or sheets list is empty, the function will clear the existing sheets or views if clear_existing is set to true. 
    Otherwise it will maintain those types of views unchanged (clear_existing=False).

    :param view_sheet_set: The view set to update.
    :type view_sheet_set: Autodesk.Revit.DB.ViewSheetSet
    :param sheets: The list of sheets to add to the view set.
    :type sheets: list
    :param views: The list of views to add to the view set.
    :type views: list
    :param clear_existing: Whether to clear existing sheets or views before adding new ones.
    :type clear_existing: bool
    :param transaction_manager: The transaction manager to use for the operation. If none is provided, the action will be executed assuming there is an open transaction already set up by the caller.
    :type transaction_manager: function

    :return: A Result object indicating the success or failure of the operation. Messages are quite detailed to allow for debugging and understanding what is happening in the function.
    :rtype: duHast.Utilities.Objects.result.Result
    """

    # set up a return object
    return_value = Result()

    # check if the view_sheet_set is None
    if not isinstance(view_sheet_set, ViewSheetSet):
        return_value.update_sep(False, "view_sheet_set needs to be of type:  ViewSheetSet.  Got instead: {}".format(type(view_sheet_set)))
        return return_value
    
    # get the ids of what is currently in the view set
    existing_sheets, existing_views = get_sheets_and_views_from_view_sheet_set(view_sheet_set)

    # store the name
    view_sheet_set_name = view_sheet_set.Name

    # check if the view_sheet_setting is None
    if view_sheet_setting is None:
        return_value.append_message("No view_sheet_setting provided, trying to get the current ViewSheetSetting element from the document.")
        view_sheet_setting = get_current_view_sheet_settings_element(doc)

    # make sure the we got the settings, otherwise we cant update
    if view_sheet_setting:
        return_value.append_message("Setting current view sheet set to: {}".format(view_sheet_set_name))
        view_sheet_setting.CurrentViewSheetSet = view_sheet_set
    else:
        return_value.update_sep(False, "Failed to get current ViewSheetSetting element.")
        return return_value

    # set up an action to be executed in a transaction
    def action():
        action_return_value = Result()
        
        try:

            # set up a new view set ( not ot be confused with view sheet set...)
            empty_set = ViewSet()
           
            # check if we have any sheets past in
            if sheets is None or len(sheets) == 0:
                action_return_value.append_message("No sheets provided")
                # if no sheets are passed in, clear existing sheets
                if not clear_existing:
                    action_return_value.append_message("adding existing sheets back to the view set")
                    # add existing sheets back to the view set
                    for sheet in existing_sheets:
                        action_return_value.append_message("Adding existing sheet: {}".format(Element.Name.GetValue(sheet)))
                        empty_set.Insert(sheet)
                else:
                    action_return_value.append_message("Not adding existing sheets back to the view set, as clear_existing is set to True")
                   
            else:
                action_return_value.append_message("Adding sheets past in to the view set")
                # if sheets are passed in, add them to the view set
                for sheet in sheets:
                    action_return_value.append_message("Adding sheet: {}".format(Element.Name.GetValue(sheet)))
                    empty_set.Insert(sheet)
            
            # check if we have any views past in
            if views is None or len(views) == 0:
                action_return_value.append_message("No views provided")
                # if no views are passed in, clear existing views
                if not clear_existing:
                    action_return_value.append_message("Adding existing views back to the view set")
                    # add existing views back to the view set
                    for view in existing_views:
                        action_return_value.append_message("Adding existing view: {}".format(Element.Name.GetValue(view)))
                        empty_set.Insert(view)
            else:
                action_return_value.append_message("Adding views past in to the view set")
                # if views are passed in, add them to the view set
                for view in views:
                    action_return_value.append_message("Adding view: {}".format(Element.Name.GetValue(view)))
                    empty_set.Insert(view)
        
            # assign the view set to the view sheet set
            action_return_value.append_message("Assigning the view set to the view sheet set: {}".format(view_sheet_set_name))
            #for element in empty_set:
            #    action_return_value.append_message("Element in view set: {}".format(Element.Name.GetValue(element)))
            
            # need to update the CurrentViewSheetSet views directly via the settings
            # otherwise update does not stick!
            view_sheet_setting.CurrentViewSheetSet.Views = empty_set

            # print the current view sheet set
            #for sheet in view_sheet_setting.CurrentViewSheetSet.Views:
            #    action_return_value.append_message("Sheet in current view sheet set: {}".format(Element.Name.GetValue(sheet)))

            # save the view sheet set
            save_flag = True
            
            if save_after_update:
                try:
                    
                    # print the current view sheet set
                    action_return_value.append_message("Current view sheet set: {}".format(view_sheet_setting.CurrentViewSheetSet.Name))
                    #for sheet in view_sheet_setting.CurrentViewSheetSet.OrderedViewList:
                    #    action_return_value.append_message("Sheet in current view sheet set: {}".format(Element.Name.GetValue(sheet)))


                    # this can throw an exception if the set is unchanged...
                    save_flag = view_sheet_setting.Save()
                except Exception as ex:
                    action_return_value.append_message("Failed to save the updated view sheet set: {} with error: {}".format(view_sheet_set_name, ex))

            # update the action return value with the result of the save operation
            action_return_value.append_message("Updated view set: {} with status: {}".format(view_sheet_set_name, save_flag))

            # return the newly created view set
            action_return_value.result.append(view_sheet_set)

        except Exception as e:
            action_return_value.update_sep(
                False,
                "Failed to update view set: {} with error: {}".format(view_sheet_set, e),
            )
        
        return action_return_value

    # check if a transaction manager is provided
    if not transaction_manager:
        # if no transaction manager is provided, run the action assuming there is an open transaction already set up by the caller
        return_value = action()
    else:
        transaction = Transaction(doc, "Updating view set {}".format(view_sheet_set_name))
        return_value = transaction_manager(transaction, action)
   
    return return_value


def update_view_sheet_set_by_name(doc, view_sheet_set_name, sheets, views, clear_existing=True, transaction_manager = in_transaction):
    """
    Updates a view set with the given name, sheets, and views in the Revit document.

    :param doc: The Revit document in which to update the view set.
    :type doc: Autodesk.Revit.DB.Document
    :param view_sheet_set_name: The name of the view set to update.
    :type view_sheet_set_name: str
    :param sheets: A list of sheets to include in the view set.
    :type sheets: list
    :param views: A list of views to include in the view set.
    :type views: list
    :param clear_existing: Whether to clear existing sheets or views before adding new ones.
    :type clear_existing: bool
    :param transaction_manager: The transaction manager to use for the operation. If none is provided, the action will be executed assuming there is an open transaction already set up by the caller.
    :type transaction_manager: function

    :return: A Result object indicating the success or failure of the operation. Messages are quite detailed to allow for debugging and understanding what is happening in the function.
                if the view set does not exist, the result will contain the name of the view set that was not found. ( that is the only time the result will contain as single string)
    :rtype: duHast.Utilities.Objects.result.Result
    """

    # set up a return object
    return_value = Result()

    try:

        # check if the view_sheet_set_name is a string
        if not isinstance(view_sheet_set_name, str):
            return_value.update_sep(False, "view_sheet_set_name needs to be of type: str. Got instead: {}".format(type(view_sheet_set_name)))
            return return_value

        if view_sheet_set_name == "":
            return_value.update_sep(False, "view_sheet_set_name cannot be an empty string.")
            return return_value

        # check view set with given name already exists?
        # if it does not, return an error
        existing_view_sheet_set = get_view_sheet_set_by_name(doc, view_sheet_set_name)

        # check if the view sheet set exists
        if not existing_view_sheet_set:
            return_value.update_sep(False, "View set with name '{}' does not exist.".format(view_sheet_set_name))
            return_value.result.append(view_sheet_set_name)
            return return_value

        # Now you can access the ViewSheetSetting
        view_sheet_setting = get_current_view_sheet_settings_element(doc)

        # check if the view_sheet_setting is None
        if not view_sheet_setting:
            return_value.update_sep(False, "Failed to get current ViewSheetSetting element.")
            return return_value
        
        # update the view set with the given sheets and views
        return_value = update_view_sheet_set(
            doc=doc, 
            view_sheet_set= existing_view_sheet_set, 
            view_sheet_setting=view_sheet_setting, 
            sheets=sheets, 
            views= views, 
            clear_existing=True, 
            transaction_manager = in_transaction, 
            save_after_update=True
        )

    except Exception as e:
        return_value.update_sep(False, "Failed to update view set with name: {} with error: {}".format(view_sheet_set_name, e))


    return return_value


def create_new_view_sheet_set(doc, view_sheet_set_name, sheets, views, transaction_manager = in_transaction):
    """
    Creates a new view set with the given name, sheets, and views in the Revit document.

    :param doc: The Revit document in which to create the view set.
    :type doc: Autodesk.Revit.DB.Document
    :param view_sheet_set_name: The name of the new view set to create.
    :type view_sheet_set_name: str
    :param sheets: A list of sheets to include in the new view set.
    :type sheets: list
    :param views: A list of views to include in the new view set.
    :type views: list
    :param transaction_manager: The transaction manager to use for the operation. If none is provided, the action will be executed assuming there is an open transaction already set up by the caller.
    :type transaction_manager: function

    :return: A Result object indicating the success or failure of the operation. Messages are quite detailed to allow for debugging and understanding what is happening in the function.
    :rtype: duHast.Utilities.Objects.result.Result
    """
    
    # set up a return object
    return_value = Result()

    # check if the view_sheet_set_name is a string
    if not isinstance(view_sheet_set_name, str):
        return_value.update_sep(False, "view_sheet_set_name needs to be of type: str. Got instead: {}".format(type(view_sheet_set_name)))
        return return_value

    if view_sheet_set_name == "":
        return_value.update_sep(False, "view_sheet_set_name cannot be an empty string.")
        return return_value

    # check view set with given name already exists?
    # if it does, return an error
    # otherwise duplicate the current set, rename it , clear it and update it with the given sheets and views
    existing_view_sheet_set = get_view_sheet_set_by_name(doc, view_sheet_set_name)

    if existing_view_sheet_set:
        return_value.update_sep(False, "View set with name '{}' already exists.".format(view_sheet_set_name))
        return return_value

    # Now you can access the ViewSheetSetting
    view_sheet_setting = get_current_view_sheet_settings_element(doc)

    # check if the view_sheet_setting is None
    if not view_sheet_setting:
        return_value.update_sep(False, "Failed to get current ViewSheetSetting element.")
        return return_value


    # set up an action to be executed in a transaction
    def action():
        action_return_value = Result()
        
        try:
            # create a new view sheet set by saving the current view sheet set with the given name
            new_view_sheet_set_flag = view_sheet_setting.SaveAs(view_sheet_set_name)

            # check if the new view set is None
            if not new_view_sheet_set_flag :
                action_return_value.update_sep(False, "Failed to create new view set.")
                return action_return_value
            
            # get the newly created view set
            new_view_sheet_set = view_sheet_setting.CurrentViewSheetSet
            
            # update the view set with the given sheets and views
            update_status = update_view_sheet_set(
                doc, 
                new_view_sheet_set,
                view_sheet_setting,
                sheets, 
                views, 
                clear_existing=True, 
                transaction_manager=None, 
                save_after_update=False
            )

            # check if successful update
            if not update_status.status:
                action_return_value.update_sep(False, "Failed to update new view set: {} with error: {}".format(view_sheet_set_name, update_status.message))
                return action_return_value

            action_return_value.update(update_status)
            save_flag = True

            try:
                # save the set
                # this can throw an exception if the set is unchanged...
                save_flag = view_sheet_setting.Save()
            except Exception as ex:
                action_return_value.append_message("Failed to save newly created view sheet set: {} with error: {} ".format(view_sheet_set_name, ex))

            # update the action return value with the result of the save operation
            action_return_value.append_message("Created new view set: {} with status: {}".format(view_sheet_set_name, save_flag))

            # return the newly created view set
            action_return_value.result.append(new_view_sheet_set)

        except Exception as e:
            action_return_value.update_sep(
                False,
                "Failed to create new view set: {} with error: {}".format(view_sheet_set_name, e),
            )
        
        return action_return_value

    # check if a transaction manager is provided
    if not transaction_manager:
        # if no transaction manager is provided, run the action assuming there is an open transaction already set up by the caller
        return_value = action()
    else:
        transaction = Transaction(doc, "Creating view set: {}".format(view_sheet_set_name))
        return_value = transaction_manager(transaction, action)
   
    return return_value


def delete_view_sheet_set_by_name(doc, view_sheet_set_name, transaction_manager = in_transaction):
    """
    Deletes a view sheet set by its name from the Revit document.

    :param doc: The Revit document from which to delete the view sheet set.
    :type doc: Autodesk.Revit.DB.Document
    :param view_sheet_set_name: The name of the view sheet set to delete.
    :type view_sheet_set_name: str
    :param transaction_manager: The transaction manager to use for the operation. If none is provided, the action will be executed assuming there is an open transaction already set up by the caller.
    :type transaction_manager: function

    :return: A Result object indicating the success or failure of the operation. Messages are quite detailed to allow for debugging and understanding what is happening in the function.
    :rtype: duHast.Utilities.Objects.result.Result
    """

    # set up a return object
    return_value = Result()

    # check if the view_sheet_set_name is a string
    if not isinstance(view_sheet_set_name, str):
        return_value.update_sep(False, "view_sheet_set_name needs to be of type: str. Got instead: {}".format(type(view_sheet_set_name)))
        return return_value

    if view_sheet_set_name == "":
        return_value.update_sep(False, "view_sheet_set_name cannot be an empty string.")
        return return_value

    view_sheet_setting = get_current_view_sheet_settings_element(doc)

    # check view set with given name already exists?
    # if it does, return an error
    # otherwise duplicate the current set, rename it , clear it and update it with the given sheets and views
    existing_view_sheet_set = get_view_sheet_set_by_name(doc, view_sheet_set_name)

    # check if the view sheet set exists
    if not existing_view_sheet_set:
        return_value.update_sep(False, "View set with name '{}' does not exist.".format(view_sheet_set_name))
        return return_value

    # make sure the we got the settings, otherwise we cant update
    if view_sheet_setting:
        return_value.append_message("Setting current view sheet set to: {}".format(view_sheet_set_name))
    else:
        return_value.update_sep(False, "Failed to get current ViewSheetSetting element.")
        return return_value
    
    # set up an action to be executed in a transaction
    def action():
        action_return_value = Result()
        
        try:
            # set the current view sheet set to the one we want to delete
            view_sheet_setting.CurrentViewSheetSet = existing_view_sheet_set
            # attempt to delete the view sheet set
            view_sheet_setting.Delete()

            action_return_value.append_message("Deleted view sheet set: {}".format(view_sheet_set_name))

        except Exception as e:
            action_return_value.update_sep(
                False,
                "Failed to delete view set: {} with error: {}".format(view_sheet_set_name, e),
            )
        
        return action_return_value

    # check if a transaction manager is provided
    if not transaction_manager:
        # if no transaction manager is provided, run the action assuming there is an open transaction already set up by the caller
        return_value = action()
    else:
        transaction = Transaction(doc, "Deleting view set: {}".format(view_sheet_set_name))
        return_value = transaction_manager(transaction, action)
   
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit symbolic curve create helper functions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

These functions work within a Revit family document to create and manipulate symbolic curves.


"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.result import Result
from duHast.Revit.Common.transaction import in_transaction
from duHast.Revit.Categories.categories import (
    ELEMENTS_PARAS_SUB
)

from duHast.Revit.Family.Geometry.utils import set_curve_sub_category

from Autodesk.Revit.DB import Curve, Transaction, SketchPlane

def create_symbolic_curves_by_curve_loop(doc, sketch_plane, curve_loop, func, transaction_manager=in_transaction):
    """
    Create new symbolic curves in the family document using the provided curve loop.
    
    
    :param doc: The family document.
    :type doc: Autodesk.Revit.DB.Document
    :param sketch_plane: The sketch plane to use for the new curve(s).
    :type sketch_plane: Autodesk.Revit.DB.SketchPlane
    :param curve_loop: The curve loop to use for the new curve(s).
    :type curve_loop:  Autodesk.Revit.DB.CurveLoop
    :param func: A function to execute on the new curves. ( need to take the doc and  a single new curves as arguments and need to return a Result class instance)
    :type func: function
    :param transaction_manager: A function to manage transactions. (default is in_transaction)
    :type transaction_manager: function
    
    :return: Result class instance.

        - `result.status` (bool): True if the symbolic curves where created successfully, otherwise False.
        - `result.message` (str): Confirmation of successful creation of the symbolic curves.
        - `result.result` (list): The new symbolic curve elements.
        
    On exception:
    
        - `result.status` (bool): False.
        - `result.message` (str): Generic exception message.
        - `result.result` (list): Empty.
        
    :rtype: :class:`.Result`
    """
    
    return_value = Result()
    try:
        
        # create new symbolic curves in the family document
        def action():
       
            action_return_value = Result()
            try:
                # Create new curves from the curve loop
                for curve in curve_loop:
                    
                    # check if the curve is a valid Curve object
                    if (not isinstance(curve, Curve)):
                        action_return_value.update_sep(False, "Element is not a valid Curve object: {}".format(type(curve)))
                        continue
                    
                    new_curve = doc.FamilyCreate.NewSymbolicCurve(curve, sketch_plane)
                
                    # check if the curve was created successfully
                    if new_curve is None:
                        action_return_value.update_sep(False, "Failed to create new curve in family")
                    else:
                        action_return_value.append_message("Created new curve in family")
                    
                    # execute function on symbolic curve
                    if func is not None:
                        func_result = func(doc, new_curve)
                        action_return_value.update(func_result)
                
                    # add the new symbolic curve to the result
                    action_return_value.result.append(new_curve)
                
            except Exception as e:
                action_return_value.update_sep(False, "Failed to create new curve(s) in family: {}".format(e))
            return action_return_value

        if transaction_manager :
            transaction = Transaction(doc, "Creating symbolic curves")
            return_value = transaction_manager(transaction,action )
        else:
            return_value = action()
            
    except Exception as e: 
        message = "Failed to create symbolic curves in family: {}".format(e)
        return_value.update_sep(False, message)
    return return_value


def create_symbolic_curves_on_level (doc, level, curve_loop, func, transaction_manager=in_transaction):
    """
    Create a new symbolic curves in the family document on the level provided using the provided curve loop.

    :param doc: The family document.
    :type doc: Autodesk.Revit.DB.Document
    :param level: The level to use for the new symbolic curves.
    :type level: Autodesk.Revit.DB.Level
    :param curve_loops: The curve loop to use for the new symbolic curves.
    :type curve_loops: Autodesk.Revit.DB.CurveLoop
    :param func: A function to execute on the new symbolic curves. ( need to take the doc and a new symbolic curve as arguments and need to return a Result class instance)
    :type func: function
    :param transaction_manager: A function to manage transactions. (default is in_transaction)
    :type transaction_manager: function

        - `result.status` (bool): True if the symbolic curves where created successfully, otherwise False.
        - `result.message` (str): Confirmation of successful creation of the symbolic curves.
        - `result.result` (list): The new symbolic curve elements.
    
    On exception:
    
        - `result.status` (bool): False.
        - `result.message` (str): Generic exception message.
        - `result.result` (list): Empty.
        
    :rtype: :class:`.Result`
    """
    
    return_value = Result()
    try:
        
        # get the reference plane of the level
        level_reference = level.GetPlaneReference()
               
        # create a new symbolic curves in the family document
        def action():
       
            action_return_value = Result()
            try:
                # Create new curves from the curve loop
                for curve in curve_loop:
                    
                    # check if the curve is a valid Curve object
                    if (not isinstance(curve, Curve)):
                        action_return_value.update_sep(False, "Element is not a valid Curve object: {}".format(type(curve)))
                        continue
                    
                    # create a sketch plane on the level reference
                    sketch_plane = SketchPlane.Create(doc, level_reference)
                    
                    # Create new symbolic curves
                    new_curve = doc.FamilyCreate.NewSymbolicCurve(curve, sketch_plane)
                    
                    # check if the symbolic curves was created successfully
                    if new_curve is None:
                        action_return_value.update_sep(False, "Failed to create new symbolic curve in family")
                    else:
                        action_return_value.append_message("Created new symbolic curve in family")
                        
                        # execute function on symbolic curves
                        if func is not None:
                            func_result = func(doc, new_curve)
                            action_return_value.update(func_result)
                    
                    # add the new symbolic curve to the result
                    action_return_value.result.append(new_curve)
                
            except Exception as e:
                action_return_value.update_sep(False, "Failed to create new symbolic curves in family: {}".format(e))
            return action_return_value

        if transaction_manager :
            transaction = Transaction(doc, "Creating symbolic curves")
            return_value = transaction_manager(transaction,action )
        else:
            return_value = action()
            
    except Exception as e: 
        message = "Failed to create symbolic curves in family: {}".format(e)
        return_value.update_sep(False, message)
    return return_value


def set_symbolic_curve_sub_category(doc, curve, source_graphic_style, transaction_manager=in_transaction):
    """
    Set the subcategory of an symbolic curve in a family document.
    The subcategory is set to the subcategory of the source projection graphic style.
    
    :param doc: The family document.
    :type doc: Autodesk.Revit.DB.Document
    :param curve: The curve element.
    :type curve: Autodesk.Revit.DB.Curve
    :param source_graphic_style: The graphic style of the source curve.
    :type source_graphic_style: Autodesk.Revit.DB.GraphicStyle
    :return: Result class instance.

        - `result.status` (bool): True if the curve subcategory was set successfully, otherwise False.
        - `result.message` (str): Confirmation of successful setting of the curve subcategory.
        - `result.result` (list): Empty.
    
    
    On exception:
        - `result.status` (bool): False.
        - `result.message` (str): Generic exception message.
        - `result.result` (list): Empty.
        
    :rtype: :class:`.Result`
    """
        
    return_value = Result()
    try:
        return_value = set_curve_sub_category(
            doc=doc, 
            element=curve, 
            source_graphic_style=source_graphic_style, 
            transaction_manager=transaction_manager,
        )
    except Exception as e:
        message = "Failed to set curve sub category in family: {}".format(e)
        return_value.update_sep(False, message)
    return return_value